"""Dataset, producer, and config metadata."""

import logging
import warnings

import sqlalchemy as sa

from .._globals import REGISTRY as registry  # noqa: N811

from .. import _tools
from .. import backend as _backend

__all__ = ['Dataset', 'Producer', 'Config']


log = logging.getLogger(__name__)


@registry.mapped
class Dataset:
    """Git commit loaded into the database."""

    __tablename__ = '__dataset__'

    id = sa.Column(sa.Integer, sa.CheckConstraint('id = 1'), primary_key=True)

    title = sa.Column(sa.Text, sa.CheckConstraint("title != ''"), nullable=False)

    git_commit = sa.Column(sa.String(40), sa.CheckConstraint('length(git_commit) = 40'),
                           nullable=False, unique=True)
    git_describe = sa.Column(sa.Text, sa.CheckConstraint("git_describe != ''"),
                             nullable=False, unique=True)
    clean = sa.Column(sa.Boolean(create_constraint=True), nullable=False)

    version = sa.Column(sa.Text, sa.CheckConstraint("version != ''"))

    exclude_raw = sa.Column(sa.Boolean(create_constraint=True), nullable=False)

    @classmethod
    def get_dataset(cls, /, *, bind, strict, fallback=None):
        table = cls.__tablename__
        log.debug('read %r from %r', table, bind)

        try:
            result, = _backend.iterrows(sa.select(cls), mappings=True, bind=bind)
        except sa.exc.OperationalError as e:
            if 'no such table' in e.orig.args[0]:
                pass
            else:
                log.exception('error selecting %r', table)
                if strict:  # pragma: no cover
                    raise RuntimeError('failed to select %r from %r', table, bind) from e
            return fallback
        except ValueError as e:
            log.exception('error selecting %r', table)
            if 'not enough values to unpack' in e.args[0] and not strict:
                return fallback
            else:  # pragma: no cover
                raise RuntimeError('failed to select %r from %r', table, bind) from e
        except Exception as e:  # pragma: no cover
            log.exception('error selecting %r', table)
            raise RuntimeError('failed to select %r from %r', table, bind) from e
        else:
            return result

    @classmethod
    def log_dataset(cls, params, /, *,
                    ignore_dirty: bool = False,
                    also_print: bool = False, print_file=None):
        name = cls.__tablename__
        log.info('git describe %(git_describe)r clean: %(clean)r', params)
        log.debug('%s.title: %r', name, params['title'])
        log.info('%s.git_commit: %r', name, params['git_commit'])
        if 'version' in params:
            log.info('%s.version: %r', name, params['version'])
        log.debug('%s.exclude_raw: %r', name, params['exclude_raw'])
        if also_print or print_file is not None:
            print('git describe {git_describe!r}'
                  ' clean: {clean!r}'.format_map(params),
                  file=print_file)
            print(f"{name}.title: {params['title']!r}'",
                  file=print_file)
            print(f"{name}.git_commit: {params['git_commit']!r}",
                  file=print_file)
            if 'version' in params:
                print(f"{name}.version: {params['version']!r}",
                      file=print_file)
            print(f"{name}.exclude_raw: {params['exclude_raw']!r}",
                  file=print_file)

        if not params['clean'] and not ignore_dirty:
            warnings.warn(f'{name} not clean,'
                          ' pass ignore_dirty=True to disable')  # pragma: no cover


@registry.mapped
class Producer:
    """Name and version of the package that created a __dataset__."""

    __tablename__ = '__producer__'

    id = sa.Column(sa.Integer, sa.CheckConstraint('id = 1'), primary_key=True)

    name = sa.Column(sa.Text, sa.CheckConstraint("name != ''"),
                     unique=True, nullable=False)

    version = sa.Column(sa.Text, sa.CheckConstraint("version != ''"),
                        nullable=False)

    @classmethod
    def get_producer(cls, /, *, bind):
        result, = _backend.iterrows(sa.select(cls), mappings=True, bind=bind)
        return result

    @classmethod
    def log_producer(cls, params, /, *,
                     also_print: bool = False, print_file=None):
        name = cls.__tablename__
        log.info('%s.name: %s', name, params['name'])
        log.info('%s.version: %s', name, params['version'])
        if also_print or print_file is not None:
            print(f"{name}.name: {params['name']}", file=print_file)
            print(f"{name}.version: {params['version']}", file=print_file)


@registry.mapped
class Config:
    """Configuration setting from ``glottolog/config/*.ini``."""

    __tablename__ = '_config'

    filename = sa.Column(sa.String, sa.CheckConstraint("filename != ''"),
                         primary_key=True)
    section = sa.Column(sa.String, sa.CheckConstraint("section != ''"),
                        primary_key=True)
    option = sa.Column(sa.String, sa.CheckConstraint("option != ''"),
                       primary_key=True)

    value = sa.Column(sa.Text, sa.CheckConstraint("value != ''"),
                      nullable=False)

    line = sa.Column(sa.Integer, sa.CheckConstraint('line > 0'),
                     nullable=False)

    __table_args__ = (sa.UniqueConstraint(filename, line),
                      {'info': {'without_rowid': True}})

    @classmethod
    def load(cls, filename: str, /, *, bind,
             _groupby_section=_tools.groupby_itemgetter(0)):
        select_values = (sa.select(Config.section, Config.option, Config.value)
                        .filter_by(filename=filename)
                        .order_by('section', 'option'))
        result = _backend.iterrows(select_values, bind=bind)
        return {section: {option: value for _, option, value in grp}
                for section, grp in _groupby_section(result)}



import sqlalchemy as sa

import pytest

INFO_KWARGS = {'memory_usage': 'deep'}


pytestmark = pytest.mark.pandas


@pytest.mark.filterwarnings('ignore'
                            ':The MetaData.bind argument is deprecated'
                            ':DeprecationWarning')
@pytest.mark.filterwarnings('ignore'
                            ':the imp module is deprecated'
                            ':DeprecationWarning')
def test_pd_read_sql(treedb):
    query = sa.select(treedb.Languoid)

    df = treedb.pd_read_sql(query, index_col='id')

    if treedb.backend.pandas.PANDAS is None:
        assert df is None
    else:
        assert not df.empty
        assert df.index.name == 'id'
        assert list(df.columns)[:5] == ['name', 'level', 'parent_id',
                                        'hid', 'iso639_3']
        assert df.index.is_unique
        df.info(**INFO_KWARGS)


def test_pd_read_json_lines(treedb):
    query = sa.select(sa.func.json_object('id', treedb.Languoid.id,
                                          'name', treedb.Languoid.name))
    df = treedb.pd_read_json_lines(query, concat_ignore_index=True)

    if treedb.backend.pandas.PANDAS is None:
        assert df is None
    else:
        assert not df.empty
        assert list(df.columns) == ['id', 'name']
        assert df.index.is_unique
        df.info(**INFO_KWARGS)


@pytest.mark.xfail(reason="broken pd.read_json(orient='index', lines=True)",
                   raises=AttributeError)
def test_pd_read_json_lines_orient_index(treedb):
    languoid = sa.func.json_object('name', treedb.Languoid.name,
                                   'level', treedb.Languoid.level)
    languoid = sa.func.json_object(treedb.Languoid.id, languoid)
    query = sa.select(languoid)

    df = treedb.pd_read_json_lines(query, orient='index')

    if treedb.backend.pandas.PANDAS is None:
        assert df is None
    else:
        assert not df.empty
        assert df.index.name == 'id'
        assert list(df.columns) == ['name', 'level']
        assert df.index.is_unique
        df.info(**INFO_KWARGS)



"""Load ``glottolog/config/*.ini`` files."""

import logging

from . import _globals
from . import _tools
from . import languoids as _languoids

__all__ = ['iterconfigs',
           'load_config']

CONFIG_IN_ROOT = _tools.path_from_filename('config')


log = logging.getLogger(__name__)


def get_config_path(root=_globals.ROOT, /, *,
                    configpath=CONFIG_IN_ROOT):
    return _languoids.get_repo_root(root) / configpath


def iterconfigs(root=_globals.ROOT, /, *, glob: str = '*.ini'):
    for path in get_config_path().glob(glob):
        yield path.name, load_config(path)


def load_config(filepath, /, *, sort_sections: bool = False
                ) -> dict[str, dict[str, str]]:
    log.debug('open config file from path: %r', filepath)
    cfg = _tools.ConfigParser.from_file(filepath)

    log.debug('parsed %d section: %r', len(cfg), cfg)
    return cfg.to_dict(sort_sections=sort_sections)



"""Check application-specific invariants."""

import inspect
import itertools
import logging

import pytest
import sqlalchemy as sa
import sqlalchemy.orm

from . import _globals
from ._globals import SESSION as Session  # noqa: N811
from .backend.models import Dataset
from .models import (FAMILY, LANGUAGE, DIALECT,
                     Languoid, PseudoFamily, Altname, AltnameProvider)

__all__ = ['check',
           'compare_languoids']


log = logging.getLogger(__name__)


def check(func=None, /, *, bind=_globals.ENGINE):
    """Run consistency/sanity checks on database."""
    if func is not None:
        try:
            check.registered.append(func)
        except AttributeError:
            check.registered = [func]
        return func

    with bind.connect() as conn:
        exclude_raw = conn.scalar(sa.select(Dataset.exclude_raw))

    passed = True

    for func in check.registered:
        with Session(bind=bind) as session:
            ns = {'invalid_query': staticmethod(func), '__doc__': func.__doc__}
            check_cls = type(f'{func.__name__}Check', (Check,), ns)

            kwargs = ({'exclude_raw': exclude_raw}
                      if func.__name__ == 'no_empty_files' else {})

            check_inst = check_cls(session, **kwargs)

            log.debug('validate %r', func.__name__)
            check_passed = check_inst.validate()

            if not check_passed:
                passed = False

    return passed


class Check(object):

    detail = True

    def __init__(self, session, /, **kwargs):
        self.session = session
        self.query = self.invalid_query(**kwargs)

    def invalid_query(self, /, **kwargs):  # pragma: no cover
        raise NotImplementedError

    def validate(self):
        query = (self.query
                 .with_only_columns(sa.func.count().label('invalid_count'))
                 .order_by(None))

        with self.session as session:
            self.invalid_count = session.execute(query).scalar()

        log.debug('invalid count: %d', self.invalid_count)
        print(self)

        if self.invalid_count:
            if self.detail:
                self.invalid = session.execute(self.query).scalars().all()
                self.show_detail(self.invalid, self.invalid_count)
            return False
        else:
            self.invalid = []
            return True

    def __str__(self):
        if self.invalid_count:
            msg = (f'{self.invalid_count:d} invalid\n'
                   f'    (violating {self.__doc__})')
        else:
            msg = 'OK'
        return f'{self.__class__.__name__}: {msg}'

    @staticmethod
    def show_detail(invalid, invalid_count, /, *, number=25):  # pragma: no cover
        ids = (i.id for i in itertools.islice(invalid, number))
        cont = ', ...' if number < invalid_count else ''
        print(f"    {', '.join(ids)}{cont}")


def docformat(func):
    sig = inspect.signature(func)
    defaults = {n: p.default for n, p in sig.parameters.items()
                if p.default != inspect.Parameter.empty}
    func.__doc__ = func.__doc__.format_map(defaults)
    return func


@check
def valid_pseudofamily_references():
    """Pseudofamilies languoid_id and name point to the same languoid."""
    return (sa.select(PseudoFamily)
            .join_from(PseudoFamily, Languoid,
                       PseudoFamily.languoid_id == Languoid.id)
            .where(PseudoFamily.name != Languoid.name))


@check
def pseudofamilies_are_roots():
    """Pseudofamilies are at the root level, i.e. parent_id is NULL."""
    return (sa.select(PseudoFamily)
            .join_from(PseudoFamily, Languoid,
                       PseudoFamily.languoid)
            .where(Languoid.parent_id != sa.null()))


@check
@docformat
def valid_glottocode(*, pattern=r'^[a-z0-9]{4}\d{4}$'):
    """Glottocodes match {pattern!r}."""
    return (sa.select(Languoid)
            .order_by('id')
            .where(~Languoid.id.regexp_match(pattern)))


@check
@docformat
def valid_iso639_3(*, pattern=r'^[a-z]{3}$'):
    """Iso codes match {pattern!r}."""
    return (sa.select(Languoid)
            .order_by('id')
            .where(~Languoid.iso639_3.regexp_match(pattern)))


@check
@docformat
def valid_hid(*, pattern=r'^(?:[a-z]{3}|NOCODE_[A-Z][a-zA-Z0-9-]+)$'):
    """Hids match {pattern!r}."""
    return (sa.select(Languoid)
            .order_by('id')
            .where(~Languoid.hid.regexp_match(pattern)))


@check
def clean_name():
    """Glottolog names lack problematic characters."""
    gl = (sa.select(AltnameProvider.id)
          .filter_by(name='glottolog')
          .scalar_subquery())

    def cond(col):
        yield col.startswith(' ')
        yield col.endswith(' ')
        yield col.regexp_match('[`_*:\xa4\xab\xb6\xbc]')  # \xa4.. common in mojibake

    match_gl = (Languoid.altnames
                .any(sa.or_(*cond(Altname.name)), provider_id=gl))

    return (sa.select(Languoid)
            .order_by('id')
            .where(sa.or_(match_gl, *cond(Languoid.name))))


@check
def family_parent():
    """Parent of a family is a family."""
    parent = sa.orm.aliased(Languoid)
    return (sa.select(Languoid)
            .filter_by(level=FAMILY)
            .order_by('id')
            .join(Languoid.parent.of_type(parent))
            .where(parent.level != FAMILY))


@check
def language_parent():
    """Parent of a language is a family."""
    parent = sa.orm.aliased(Languoid)
    return (sa.select(Languoid)
            .filter_by(level=LANGUAGE)
            .order_by('id')
            .join(Languoid.parent.of_type(parent))
            .where(parent.level != FAMILY))


@check
def dialect_parent():
    """Parent of a dialect is a language or dialect."""
    parent = sa.orm.aliased(Languoid)
    return (sa.select(Languoid)
            .filter_by(level=DIALECT)
            .order_by('id')
            .join(Languoid.parent.of_type(parent))
            .where(parent.level.notin_([LANGUAGE, DIALECT])))


@check
def family_children():
    """Family has at least one subfamily or language."""
    return (sa.select(Languoid)
            .filter_by(level=FAMILY)
            .order_by('id')
            .where(~Languoid.children.any(Languoid.level.in_([FAMILY,
                                                              LANGUAGE]))))


@check
def family_languages():
    """Family has at least two languages (except 'Unclassified ...')."""
    Family, Child = (sa.orm.aliased(Languoid, name=n)  # noqa: N806
                     for n in ('family', 'child'))

    tree = Languoid.tree(include_self=False, with_terminal=True)

    return (sa.select(Languoid)
            .filter_by(level=FAMILY)
            .order_by('id')
            .where(~Languoid.name.startswith('Unclassified '))
            .where(~sa.select(Family)
                   .filter_by(level=FAMILY)
                   .join(PseudoFamily, Family.pseudofamily)
                   .join(tree, Family.id == tree.c.parent_id)
                   .filter_by(child_id=Languoid.id, terminal=True)
                   .exists())
            .where(sa.select(sa.func.count())
                   .select_from(Child)
                   .filter_by(level=LANGUAGE)
                   .join(tree, Child.id == tree.c.child_id)
                   .filter_by(parent_id=Languoid.id)
                   .scalar_subquery() < 2))


def bookkeeping_no_children():
    """Bookkeeping languoids lack children (book1242 is flat).

    changed in Glottolog 4.4.1.dev0
    https://github.com/glottolog/glottolog/pull/78
    see e.g. book1242/pend1244 in 3c83d9f685f230f5cb1880e278536793c71cc69e
    """
    parent = sa.orm.aliased(Languoid)
    return (sa.select(Languoid)
            .select_from(Languoid)
            .order_by('id')
            .where(sa.exists()
                   .where(Languoid.parent_id == parent.id)
                   .where(parent.id == PseudoFamily.languoid_id)
                   .where(PseudoFamily.bookkeeping))
            .where(Languoid.children.any()))


@check
def no_empty_files(*, exclude_raw: bool):
    if exclude_raw:  # pragma: no cover
        pytest.skip('skipped from exclude_raw=True')
        return sa.select(sa.true()).where(sa.false())

    from .raw import File, Value

    return (sa.select(File)
            .select_from(File)
            .where(~sa.exists().where(Value.file_id == File.id)))


def compare_languoids(left_source: str = 'files', right_source: str = 'raw', /,
                      *, order_by: str = _globals.LANGUOID_ORDER):  # pragma: no cover
    from . import export

    def compare(left, right):
        same = True
        for lt, rt in itertools.zip_longest(left, right):
            if lt != rt:
                same = False
                print('', '', lt, '', rt, '', '', sep='\n')

        return same

    left, right = (export.iterlanguoids(source, order_by=order_by)
                   for source in (left_source, right_source))

    return compare(left, right)



"""``pytest`` command-line options and fixtures."""

import os
import string

import pytest

DEFAULT_GLOTTOLOG_TAG = 'v5.3'

RUN_WRITES = '--run-writes'

SKIP_SLOW = '--skip-slow'

SKIP_PANDAS = '--skip-pandas'

SKIP_SQLPARSE = '--skip-sqlparse'

GLOTTOLOG_TAG = '--glottolog-tag'

EXCLUDE_RAW = '--exclude-raw'


def pytest_addoption(parser):
    parser.addoption(RUN_WRITES, action='store_true',
                     help='run tests with pytest.mark.writes')

    parser.addoption(SKIP_SLOW, action='store_true',
                     help='skip tests with pytest.mark.slow')

    parser.addoption(SKIP_PANDAS, action='store_true',
                     help='skip tests with optional pandas dependency')

    parser.addoption(SKIP_SQLPARSE, action='store_true',
                     help='skip tests with optional sqlparse dependency')

    parser.addoption('--file-engine', action='store_true',
                     help='use configured file engine instead of in-memory db')

    parser.addoption('--file-engine-tag', metavar='TMPL', default='-$glottolog_tag',
                     help='filename extra tag template rendered with pytestconfig.option')

    parser.addoption(GLOTTOLOG_TAG, metavar='GIT_TAG', default=DEFAULT_GLOTTOLOG_TAG,
                     help='tag or branch to clone/checkout from Glottolog master repo')

    parser.addoption('--glottolog-repo-root', metavar='PATH',
                     help='pass root=PATH to treedb.configure()')

    parser.addoption('--rebuild', action='store_true',
                     help='pass rebuild=True to treedb.load()')

    parser.addoption('--force-rebuild', action='store_true',
                     help='pass force_rebuild=True to treedb.load()')

    parser.addoption(EXCLUDE_RAW, action='store_true',
                     help='pass exclude_raw=True to treedb.load()')

    parser.addoption('--loglevel-debug', action='store_true',
                     help='pass loglevel=DEBUG to treedb.configure()')

    parser.addoption('--log-sql', action='store_true',
                     help='pass log_sql=True to treedb.configure()')

    parser.addoption('--no-sqlalchemy-warn-20', action='store_false',
                     dest='sqlalchemy_warn_20',
                     help="don't set os.environ['SQLALCHEMY_WARN_20']")


def pytest_configure(config):
    file_engine_tag = string.Template(config.option.file_engine_tag)
    file_engine_tag = file_engine_tag.substitute(config.option.__dict__)
    config.option.file_engine_tag = file_engine_tag

    config.addinivalue_line('markers', f'writes: skip unless {RUN_WRITES} is given')
    config.addinivalue_line('markers', f'slow: skip if {SKIP_SLOW} flag is given')
    config.addinivalue_line('markers', f'pandas: skip if {SKIP_PANDAS} flag is given')
    config.addinivalue_line('markers', f'sqlparse: skip if {SKIP_SQLPARSE} flag is given')

    config.addinivalue_line('markers', f'skipif_glottolog_tag: skip for specific {GLOTTOLOG_TAG}')
    config.addinivalue_line('markers', f'xfail_glottolog_tag: xfail for given {GLOTTOLOG_TAG}')

    config.addinivalue_line('markers', f'raw: skip if {EXCLUDE_RAW} flag is given')

    if config.option.sqlalchemy_warn_20:
        os.environ['SQLALCHEMY_WARN_20'] = 'true'


def pytest_collection_modifyitems(config, items):
    def itermarkers():
        for keyword, option in {'writes': RUN_WRITES,
                                'slow': SKIP_SLOW,
                                'pandas': SKIP_PANDAS,
                                'sqlparse': SKIP_SQLPARSE,
                                'raw': EXCLUDE_RAW}.items():
            skip = not option.startswith('--run-')
            if config.getoption(option) == skip:
                reason = 'require' if not skip else 'skipped by'
                marker = pytest.mark.skip(reason=f'{reason} {option} flag')
                yield keyword, marker

    keyword_markers = dict(itermarkers())

    for item in items:
        for keyword, marker in keyword_markers.items():
            if keyword in item.keywords:
                item.add_marker(marker)
        for option in ('skipif_glottolog_tag', 'xfail_glottolog_tag'):
            for marker in item.iter_markers(option):
                if config.getoption(GLOTTOLOG_TAG) in marker.args:
                    marker_cls = {'skipif': pytest.mark.skip,
                                  'xfail': pytest.mark.xfail}[option.partition('_')[0]]
                    item.add_marker(marker_cls(**marker.kwargs))


def get_configure_kwargs(pytestconfig, *, title: str):
    memory_tag = '' if pytestconfig.option.file_engine else '-memory'
    title = f'{title}{memory_tag}{pytestconfig.option.file_engine_tag}'
    kwargs = {'title': title,
              'title_memory_tag': '',
              'engine': f'{title}.sqlite3'
                        if pytestconfig.option.file_engine else None,
              'loglevel': 'DEBUG' if pytestconfig.option.loglevel_debug else 'WARNING'}

    if pytestconfig.option.glottolog_repo_root is not None:
        kwargs['root'] = pytestconfig.option.glottolog_repo_root

    if pytestconfig.option.log_sql:
        kwargs['log_sql'] = True

    return kwargs


@pytest.fixture(scope='session')
def bare_treedb(pytestconfig):
    import treedb as bare_treedb

    kwargs = get_configure_kwargs(pytestconfig, title=f'{bare_treedb.__title__}-bare')
    bare_treedb.configure(**kwargs)

    bare_treedb.checkout_or_clone(pytestconfig.option.glottolog_tag)

    return bare_treedb


@pytest.fixture(scope='session')
def empty_treedb(pytestconfig, bare_treedb):
    empty_treedb = bare_treedb

    kwargs = get_configure_kwargs(pytestconfig, title=f'{empty_treedb.__title__}-empty')
    empty_treedb.configure(**kwargs)

    empty_treedb.load(_only_create_tables=True,
                      rebuild=pytestconfig.option.rebuild,
                      force_rebuild=pytestconfig.option.force_rebuild,
                      exclude_raw=pytestconfig.option.exclude_raw)

    return empty_treedb


@pytest.fixture(scope='session')
def treedb(pytestconfig, bare_treedb):
    treedb = bare_treedb

    kwargs = get_configure_kwargs(pytestconfig, title=treedb.__title__)
    treedb.configure(**kwargs)

    treedb.load(rebuild=pytestconfig.option.rebuild,
                force_rebuild=pytestconfig.option.force_rebuild,
                exclude_raw=pytestconfig.option.exclude_raw)

    return treedb


@pytest.fixture(scope='session')
def treedb_raw(treedb):
    import treedb.raw

    return treedb



"""Write information to stdout, csv, files."""

import logging
import warnings

import csv23

import sqlalchemy as sa

from .. import _globals
from .. import _tools
from ..backend import export as _backend_export
from ..languoids import files as _files

from . import records as _records

from .models import File, Option, Value

__all__ = ['print_stats',
           'checksum',
           'write_raw_csv',
           'write_files']


log = logging.getLogger(__name__)


def print_stats(*, file=None):
    log.info('fetch statistics')

    # order by descending frequency for any_options and undefined options
    select_nvalues = (sa.select(Option.section, Option.option,
                                sa.func.count().label('n'))
                      .join_from(Option, Value)
                      .group_by(Option.section, Option.option)
                      .order_by(sa.desc('defined'),
                                'ord_section', 'ord_option',
                                'section', sa.desc('n'), 'option'))

    template = '{section:<22} {option:<22} {n:,}'

    _backend_export.print_rows(select_nvalues, format_=template,
                               file=file)


def checksum(*, weak: bool = False,
             hash_name: str = _globals.DEFAULT_HASH,
             dialect: str = csv23.DIALECT,
             encoding: str = csv23.ENCODING):
    kind = {True: 'weak', False: 'strong', 'unordered': 'unordered'}[weak]
    log.info('calculate %r raw checksum', kind)

    if weak:
        select_rows = (sa.select(File.path,
                                 Option.section, Option.option,
                                 Value.value)
                       .join_from(File, Value).join(Option))

        order = ['path', 'section', 'option']
        if weak == 'unordered':
            order.append(Value.value)
        else:
            order.append(Value.line)
        select_rows = select_rows.order_by(*order)

    else:
        select_rows = (sa.select(File.path, File.sha256)
                       .order_by('path'))

    hashobj = _backend_export.hash_csv(select_rows,
                                       hash_name=hash_name,
                                       dialect=dialect, encoding=encoding,
                                       raw=True)

    logging.info('%s: %r', hashobj.name, hashobj.hexdigest())
    return f'{kind}:{hashobj.name}:{hashobj.hexdigest()}'


def write_raw_csv(filename=None, /, *,
                  dialect: str = csv23.DIALECT, encoding: str = csv23.ENCODING):
    """Write (path, section, option, line, value) rows to filename."""
    if filename is None:
        filename = _globals.ENGINE.file_with_suffix('.raw.csv.gz').name
    else:
        filename = _tools.path_from_filename(filename)

    path = _tools.path_from_filename(filename)
    if path.exists():
        warnings.warn(f'deltete present file: {path!r}')
        path.unlink()

    select_values = (sa.select(File.path,
                               Option.section, Option.option,
                               Value.line, Value.value)
                     .join_from(File, Value).join(Option)
                     .order_by('path', 'section', 'option', 'line'))

    return _backend_export.write_csv(select_values, filename,
                                     dialect=dialect, encoding=encoding)


def write_files(root=_globals.ROOT, /, *, replace: bool = False,
                dry_run: bool = False,
                require_nwritten: int | None = None,
                limit: int | None = None,
                offset: int | None = 0,
                progress_after: int = _tools.PROGRESS_AFTER,
                bind=_globals.ENGINE):
    """Write (path, section, option, line, value) rows back into config files."""
    log.info('write from raw record lines to tree')

    records = _records.fetch_records(order_by='path',
                                     progress_after=progress_after,
                                     bind=bind)

    records = _tools.islice_limit(records,
                                  limit=limit,
                                  offset=offset)

    return _files.write_files(records, root=root, replace=replace,
                              dry_run=dry_run,
                              require_nwritten=require_nwritten,
                              progress_after=progress_after)



"""Load Glottolog lanuoid tree ``md.ini`` files into SQLite3 database."""

from ._globals import SESSION as Session  # noqa: N811
from ._tools import sha256sum
from .backend import (print_versions,
                      set_engine,
                      connect,
                      scalar,
                      iterrows)
from .backend.export import (print_dataset,
                             print_schema,
                             print_query_sql,
                             backup,
                             dump_sql,
                             csv_zipfile,
                             print_rows,
                             write_csv,
                             hash_csv)
from .backend.load import main as load
from .backend.models import Dataset, Producer, Config
from .backend.pandas import pd_read_sql, pd_read_json_lines
from .backend.sqlite_master import print_table_sql, select_tables_nrows
from .backend.views import TABLES as views  # noqa: N811
from .languoids import set_root, iterfiles
from .checks import check, compare_languoids
from .export import (print_languoid_stats,
                     iterlanguoids,
                     checksum,
                     write_json_lines as write_languoids,
                     pd_read_languoids,
                     write_files)
from .glottolog import glottolog_version, checkout_or_clone
from .logging_ import configure_logging
from .models import LEVEL, Languoid
from .queries import (get_example_query,
                      get_json_query as get_languoids_query,
                      iterdescendants)
from .settings import configure, get_default_root

__all__ = ['Session',
           'sha256sum',
           'print_versions',
           'set_engine', 'connect', 'scalar', 'iterrows',
           'print_dataset',
           'print_schema', 'print_query_sql',
           'backup', 'dump_sql', 'csv_zipfile',
           'print_rows', 'write_csv', 'hash_csv',
           'load',
           'Dataset', 'Producer', 'Config',
           'pd_read_sql', 'pd_read_json_lines',
           'print_table_sql', 'select_tables_nrows',
           'views',
           'set_root', 'iterfiles',
           'check', 'compare_languoids',
           'print_languoid_stats',
           'iterlanguoids',
           'checksum',
           'write_languoids',
           'pd_read_languoids',
           'write_files',
           'glottolog_version', 'checkout_or_clone',
           'configure_logging',
           'LEVEL', 'Languoid',
           'get_example_query',
           'get_languoids_query',
           'iterdescendants',
           'configure',
           'engine', 'root']

__title__ = 'treedb'
__version__ = '2.7.2'
__author__ = 'Sebastian Bank <sebastian.bank@uni-leipzig.de>'
__license__ = 'MIT, see LICENSE.txt'
__copyright__ = 'Copyright (c) 2017-2026 Sebastian Bank'


# default engine: in-memory database
engine = set_engine(None)


# default root: GLOTTOLOG_REPO_ROOT, or treedb.ini glottolog:repo_root, or ./glottolog
root = set_root(get_default_root(env_var='GLOTTOLOG_REPO_ROOT'))



"""Clone/checkout the Glottolog master repo."""

from __future__ import annotations

import argparse
import logging

from . import _globals
from . import _tools
from . import languoids as _languoids

__all__ = ['glottolog_version',
           'checkout_or_clone',
           'git_rev_parse', 'git_describe',
           'git_status', 'git_status_is_clean']

REPO_URL = 'https://github.com/glottolog/glottolog.git'


log = logging.getLogger(__name__)


def glottolog_version(*, root=_globals.ROOT) -> argparse.Namespace:
    return GlottologVersion.from_root(root)


class GlottologVersion(argparse.Namespace):

    @classmethod
    def from_root(cls, root, /) -> GlottologVersion:
        return cls.from_commit_describe(git_rev_parse(root),
                                        git_describe(root))

    @classmethod
    def from_commit_describe(cls, commit: str, describe: str, /,
                             ) -> GlottologVersion:
        return cls(commit=commit, describe=describe)

    def __str__(self) -> str:
        return f'Glottolog {self.describe} ({self.commit})'


def checkout_or_clone(tag_or_branch: str, /, *, target=None):
    if target is None:
        target = _languoids.get_repo_root()

    target = _tools.path_from_filename(target)

    if not target.exists():
        clone = git_clone(tag_or_branch, target=target)
    else:
        clone = None

    checkout = git_checkout(tag_or_branch, target=target)
    return clone, checkout


def git_clone(tag_or_branch: str, /, *, target,
              depth: int = 1):
    log.info('clone Glottolog master repo at %r into %r', tag_or_branch, target)
    cmd = ['git', 'clone',
           '-c', 'advice.detachedHead=false',
           '--single-branch',
           '--branch', tag_or_branch,
           '--depth', f'{depth:d}',
           REPO_URL, target]
    return _tools.run(cmd, check=True)


def git_checkout(tag_or_branch: str, /, *, target,
                 set_branch: str = __package__):
    log.info('checkout %r and (re)set branch %r', tag_or_branch, set_branch)
    cmd = ['git', 'checkout']
    if set_branch is not None:
        cmd += ['-B', set_branch]
    cmd.append(tag_or_branch)
    return _tools.run(cmd, cwd=target, check=True)


def git_rev_parse(repo_root, /, *, revision: str = 'HEAD',
                  verify: bool = True) -> str:
    log.info('get %r git_commit from %r', revision, repo_root)
    cmd = ['git', 'rev-parse']
    if verify:
        cmd.append('--verify')
    cmd.append(revision)
    commit = _tools.run(cmd, cwd=repo_root, check=True,
                        capture_output=True, unpack=True)
    log.info('git_commit: %r', commit)
    return commit


def git_describe(repo_root, /) -> str:
    log.info('get git_describe from %r', repo_root)
    cmd = ['git', 'describe', '--tags', '--always']
    describe = _tools.run(cmd, cwd=repo_root, check=True,
                          capture_output=True, unpack=True)
    log.info('git_describe: %r', describe)
    return describe


def git_status(repo_root, /) -> str:
    log.debug('get status from %r', repo_root)
    cmd = ['git', 'status', '--porcelain']
    return _tools.run(cmd, cwd=repo_root, check=True,
                      capture_output=True, unpack=True)


def git_status_is_clean(repo_root, /) -> bool:
    """Return if there are neither changes in the index nor untracked files."""
    log.info('get clean from %r', repo_root)
    status = git_status(repo_root)
    clean = not status
    log.info('clean: %r', clean)
    return clean



"""SQLite3 database sub-module level globals."""

import contextlib
import logging
import sqlite3

import csv23

import pycountry
import sqlalchemy as sa
import sqlalchemy.ext.compiler

from .._globals import DEFAULT_FILESTEM, ENGINE
from .. import _globals
from .. import _proxies
from .. import _tools
from .. import logging_

from . import sqlparse

__all__ = ['print_versions',
           'set_engine',
           'connect',
           'scalar',
           'iterrows',
           'expression_compile',
           'json_object',
           'json_datetime']


log = logging.getLogger(__name__)


@sa.event.listens_for(sa.engine.Engine, 'connect')
def sqlite_engine_connect(dbapi_conn, connection_record):
    """Activate sqlite3 forein key checks."""
    if not isinstance(dbapi_conn, sqlite3.Connection):  # pragma: no cover
        return

    log.debug('connect sqlalchemy.engine.Engine: enable foreign keys')

    with contextlib.closing(dbapi_conn.cursor()) as dbapi_cursor:
        dbapi_cursor.execute('PRAGMA foreign_keys = ON')


@sa.ext.compiler.compiles(sa.schema.CreateTable)
def compile(element, compiler, **kwargs):
    """Append sqlite3 WITHOUT_ROWID to CREATE_TABLE if configured.

    From https://gist.github.com/chaoflow/3a6dc9d42a90c38870b8d4033b58a4d1
    """
    text = compiler.visit_create_table(element, **kwargs)
    if element.element.info.get('without_rowid'):
        text = text.rstrip() + ' WITHOUT ROWID'
    return text


def print_versions(*, engine=ENGINE, file=None) -> None:
    logging_.log_version(also_print=True, print_file=file)
    log_versions(also_print=True, print_file=file,
                 engine=engine)


def set_engine(filename, /, *,
               resolve: bool = False,
               require: bool = False,
               title: str | None = None,
               title_memory_tag: str = _globals.MEMORY_TAG):
    """Return new sqlite3 engine and set it as default engine for treedb."""
    log.info('set_engine: %r', filename)

    if isinstance(filename, sa.engine.Engine):
        engine = filename
        if isinstance(filename, _proxies.EngineProxy):
            engine = engine.engine

        ENGINE.engine = engine
        return ENGINE

    if filename is None:
        if title is None:
            title = DEFAULT_FILESTEM
        ENGINE.memory_write_path = _tools.path_from_filename(f'{title}{title_memory_tag}',
                                                             expanduser=False)
    else:
        del title

        filename = _tools.path_from_filename(filename)
        if resolve:
            filename = filename.resolve(strict=False)

        if require and not filename.exists():
            log.error('required engine file not found: %r', filename)
            raise RuntimeError(f'engine file does not exist: {filename!r}')

    ENGINE.file = filename
    log_versions(engine=ENGINE)
    return ENGINE


def log_versions(*, also_print=False, print_file=None,
                 engine=ENGINE):
    log.info('pycountry version: %s', pycountry.__version__)
    log.info('sqlalchemy version: %s', sa.__version__)
    log.info('sqlite version: %s', engine.dialect.dbapi.sqlite_version)
    log.info('csv23 version: %s', csv23.__version__)
    if also_print or print_file is not None:
        print(f'pycountry version: {pycountry.__version__}',
              file=print_file)
        print(f'sqlalchemy version: {sa.__version__}',
              file=print_file)
        print(f'sqlite_version: {engine.dialect.dbapi.sqlite_version}',
              file=print_file)
        print(f'csv23 version: {csv23.__version__}',
              file=print_file)
    if sqlparse is not None:  # pragma: no cover
        log.info('sqlparse version: %s', sqlparse.__version__)
        if also_print or print_file is not None:
            print(f'sqlparse version: {sqlparse.__version__}',
                  file=print_file)


def connect(*, bind=ENGINE,
            pragma_bulk_insert: bool = False,
            page_size: int | None = None):
    """Connect, log, apply SQLite insert optimization, return connection."""
    if isinstance(bind, sa.engine.base.Connection):
        assert not pragma_bulk_insert
        assert page_size is None

        log.debug('nested connect (no-op): %r', bind)
        return contextlib.nullcontext(bind)

    log.debug('engine connect: %r', bind)
    conn = bind.connect()
    log.debug('conn: %r', conn)

    dbapi_conn = conn.connection.driver_connection
    log.debug('dbapi_conn: %r', dbapi_conn)

    if page_size is not None:
        conn.execute(sa.text(f'PRAGMA page_size = {page_size:d}'))

    if pragma_bulk_insert:
        conn.execute(sa.text('PRAGMA synchronous = OFF'))
        conn.execute(sa.text('PRAGMA journal_mode = MEMORY'))

    return conn


def scalar(statement, /, *args, bind=ENGINE, **kwargs):
    with connect(bind=bind) as conn:
        return conn.scalar(statement, *args, **kwargs)


def iterrows(query, /, *, mappings=False, bind=ENGINE):
    with connect(bind=bind) as conn:
        result = conn.execute(query)

        if mappings:
            result = result.mappings()

        yield from result


def expression_compile(expression, /, *, literal_binds=True):
    """Return literal compiled expression."""
    return expression.compile(compile_kwargs={'literal_binds': literal_binds})


def json_object(*, sort_keys_: bool,
                label_: str | None = None,
                load_json_: bool = False, **kwargs):
    items = sorted(kwargs.items()) if sort_keys_ else kwargs.items()
    obj = sa.func.json_object(*[x for kv in items for x in kv])
    if label_ is not None:
        obj = obj.label(label_)
    return sa.type_coerce(obj, sa.JSON) if load_json_ else obj


def json_datetime(date, /):
    date = sa.func.replace(date, ' ', 'T')
    return sa.func.replace(date, '.000000', '')



"""Optional pandas dependency convenience short-cut functions."""

from collections.abc import Iterable
import functools
import io
import logging
import warnings

from .. import _globals
from .. import _tools
from .. import backend as _backend

__all__ = ['pd_read_sql',
           'pd_read_json_lines']

PANDAS = None

JSON_BUFLINES = 5_000


log = logging.getLogger(__name__)


def _import_pandas():
    global PANDAS

    if PANDAS is None:
        try:
            import pandas as PANDAS  # noqa: N812
        except ImportError as e:  # pragma: no cover
            warnings.warn(f'failed to import pandas: {e}')
        else:
            log.info('pandas version: %s', PANDAS.__version__)


def pd_read_sql(sql=None, /, *args, con=_globals.ENGINE, **kwargs):
    _import_pandas()

    if PANDAS is None:
        return None

    if sql is None:
        from .. import queries

        sql = queries.get_example_query()

    with _backend.connect(bind=con) as conn:
        return PANDAS.read_sql_query(sql, *args, con=conn, **kwargs)


def pd_read_json_lines(query, /, *,
                       buflines: int = JSON_BUFLINES,
                       bind=_globals.ENGINE,
                       **kwargs):
    _import_pandas()

    if PANDAS is None:
        return None

    with _backend.connect(bind=bind) as conn:
        result = conn.execute(query)
        json_lines = result.scalars()
        return _pd_read_json_lines(json_lines, **kwargs)


def _pd_read_json_lines(json_lines: Iterable[str], /, *,
                        buflines: int = JSON_BUFLINES,
                        concat_ignore_index: bool = False,
                        **kwargs):
    _import_pandas()

    if PANDAS is None:
        return None

    with io.StringIO() as buf:
        print_line = functools.partial(print, file=buf)
        df = None
        for chunk in _tools.iterslices(json_lines, size=buflines):
            for line in chunk:
                print_line(line)
            buf.seek(0)

            lines_df = PANDAS.read_json(buf, lines=True, **kwargs)
            buf.seek(0)
            buf.truncate()

            if df is None:
                df = lines_df
            else:
                df = PANDAS.concat([df, lines_df],
                                   ignore_index=concat_ignore_index,
                                   copy=False)

    return df



"""Main ``sqlalchemy`` schema for SQLite3 database."""

import sqlalchemy as sa

from sqlalchemy import (Table, Column, ForeignKey, CheckConstraint,
                        Integer, Float, String, Text, Enum, DateTime, Date,
                        UniqueConstraint, Index)

from sqlalchemy.orm import relationship, aliased

from . import _globals
from ._globals import REGISTRY as registry  # noqa: N811
from .backend import json_object, json_datetime

__all__ = ['LEVEL', 'Languoid']

FAMILY, LANGUAGE, DIALECT = LEVEL = ('family', 'language', 'dialect')

SPECIAL_FAMILIES = ('Sign Language',
                    'Unclassifiable',
                    'Pidgin',
                    'Unattested',
                    'Artificial Language',
                    'Mixed Language',
                    'Speech Register')

BOOKKEEPING = 'Bookkeeping'

LINK_SCHEME = {'https', 'http'}

SOURCE_PROVIDER = {'glottolog'}

ALTNAME_PROVIDER = {'multitree', 'lexvo', 'hhbib_lgcode',
                    'wals', 'wals other',
                    'moseley & asher (1994)', 'ruhlen (1987)',
                    'glottolog', 'ethnologue', 'elcat', 'aiatsis'}

TRIGGER_FIELD = {'lgcode', 'inlg'}

IDENTIFIER_SITE = {'multitree', 'endangeredlanguages',
                   'wals', 'languagelandscape'}

_SUB, _FAMILY = 'sub', 'family'

CLASSIFICATION = {'sub': (False, _SUB), 'subrefs': (True, _SUB),
                  'family': (False, _FAMILY), 'familyrefs': (True, _FAMILY)}

CLASSIFICATION_KIND = {c for _, c in CLASSIFICATION.values()}

EL_COMMENT_TYPE = {'Missing', 'Spurious'}

ISORETIREMENT_REASON = {'split', 'merge', 'duplicate', 'non-existent', 'change'}


@registry.mapped
class Languoid:

    __tablename__ = 'languoid'

    id = Column(String(8), CheckConstraint('length(id) = 8'), primary_key=True)

    name = Column(String, CheckConstraint("name != ''"),
                  nullable=False, unique=True)

    level = Column(ForeignKey('languoidlevel.name'), nullable=False)

    parent_id = Column(ForeignKey('languoid.id',
                                  # do not require to insert parent before child
                                  deferrable=True, initially='DEFERRED'),
                       index=True)

    hid = Column(Text, CheckConstraint('length(hid) >= 3'), unique=True)
    iso639_3 = Column(String(3), CheckConstraint('length(iso639_3) = 3'),
                      unique=True)

    latitude = Column(Float, CheckConstraint('latitude BETWEEN -90 AND 90'))
    longitude = Column(Float, CheckConstraint('longitude BETWEEN -180 AND 180'))

    __table_args__ = (CheckConstraint('(latitude IS NULL) = (longitude IS NULL)'),
                      {'info': {'without_rowid': True}})

    def __repr__(self):
        hid_iso = [f'{n}={getattr(self, n)!r}' for n in ('hid', 'iso639_3') if getattr(self, n)]
        hid_iso = ' '.join(hid_iso) if hid_iso else ''
        return (f'<{self.__class__.__name__}'
                f' id={self.id!r}'
                f' level={self.level!r}'
                f' name={self.name!r}'
                f'{hid_iso}>')

    languoidlevel = relationship('LanguoidLevel',
                                 back_populates='languoids')

    pseudofamily = relationship('PseudoFamily',
                                primaryjoin='PseudoFamily.languoid_id == Languoid.id',
                                back_populates='languoid')

    parent = relationship('Languoid', remote_side=[id])

    children = relationship('Languoid', remote_side=[parent_id], order_by=id,
                            viewonly=True)

    macroareas = relationship('Macroarea',
                              secondary='languoid_macroarea',
                              order_by='Macroarea.name',
                              back_populates='languoids')

    countries = relationship('Country',
                             secondary='languoid_country',
                             order_by='Country.id',
                             back_populates='languoids')

    links = relationship('Link',
                         order_by='Link.ord',
                         back_populates='languoid')

    timespan = relationship('Timespan',
                            back_populates='languoid')

    sources = relationship('Source',
                           order_by='[Source.provider_id, Source.bibitem_id]',
                           back_populates='languoid')

    altnames = relationship('Altname',
                            order_by='[Altname.provider_id, Altname.name, Altname.lang]',
                            back_populates='languoid')

    triggers = relationship('Trigger',
                            order_by='[Trigger.field, Trigger.ord]',
                            back_populates='languoid')

    identifiers = relationship('Identifier',
                               order_by='Identifier.site_id',
                               back_populates='languoid')

    subclassificationcomment = relationship('ClassificationComment', uselist=False,
        primaryjoin="and_(ClassificationComment.languoid_id == Languoid.id,"
                    " ClassificationComment.kind == 'sub')",
        viewonly=True)

    subclassificationrefs = relationship('ClassificationRef', order_by='ClassificationRef.ord',
        primaryjoin="and_(ClassificationRef.languoid_id == Languoid.id,"
                    "ClassificationRef.kind == 'sub')",
        viewonly=True)

    familyclassificationcomment = relationship('ClassificationComment', uselist=False,
        primaryjoin="and_(ClassificationComment.languoid_id == Languoid.id,"
                    " ClassificationComment.kind == 'family')",
        viewonly=True)

    familyclassificationrefs = relationship('ClassificationRef', order_by='ClassificationRef.ord',
        primaryjoin="and_(ClassificationRef.languoid_id == Languoid.id,"
                    "ClassificationRef.kind == 'family')",
        viewonly=True)

    endangerment = relationship('Endangerment',
                                uselist=False,
                                back_populates='languoid')

    ethnologue_comment = relationship('EthnologueComment',
                                      uselist=False,
                                      back_populates='languoid')

    iso_retirement = relationship('IsoRetirement',
                                  uselist=False,
                                  back_populates='languoid')

    @classmethod
    def _aliased_child_parent(cls, /, *,
                              child_root: bool = False,
                              parent_root: bool = False):
        if child_root and parent_root:  # pragma: no cover
            raise ValueError('child_root and parent_root are mutually exclusive')

        Child = aliased(cls, name='root' if child_root else 'child')  # noqa: N806
        Parent = aliased(cls, name='root' if parent_root else 'parent')  # noqa: N806
        return Child, Parent

    @classmethod
    def _tree(cls, /, *, from_parent: bool = False,
              innerjoin=False,
              child_root=None, parent_root=None, node_level=None,
              with_steps: bool = False,
              with_terminal: bool = False):
        if innerjoin not in (False, True, 'reflexive'):  # pragma: no cover
            raise ValueError(f'invalid innerjoin: {innerjoin!r}')

        Child, Parent = cls._aliased_child_parent(child_root=child_root,  # noqa: N806
                                                  parent_root=parent_root)

        if from_parent:
            Node, Relative = Parent, Child  # noqa: N806
            node_label, relative_label = 'parent_id', 'child_id'
            join_source, join_target = Node.id, Relative.parent_id
            recurse_relative = Relative.id
        else:
            Node, Relative = Child, Parent  # noqa: N806
            node_label, relative_label = 'child_id', 'parent_id'
            join_source, join_target = Node.parent_id, Relative.id
            recurse_relative = Relative.parent_id

        tree_1 = sa.select(Node.id.label(node_label))

        if innerjoin == 'reflexive':
            tree_1_relative = Node
        else:
            tree_1_relative = Relative
            tree_1 = tree_1.join_from(Node, Relative,
                                      join_source == join_target,
                                      isouter=not innerjoin)

        tree_1 = tree_1.add_columns(tree_1_relative.id.label(relative_label))

        if with_steps:
            steps = 0 if innerjoin == 'reflexive' else 1
            tree_1 = tree_1.add_columns(sa.literal(steps).label('steps'))

        if with_terminal:
            if from_parent:  # pragma: no cover
                raise NotImplementedError
            tree_1_terminal = Node if innerjoin == 'reflexive' else Relative
            terminal = sa.type_coerce(tree_1_terminal.parent_id == sa.null(),
                                      sa.Boolean)
            tree_1 = tree_1.add_columns(terminal.label('terminal'))

        if child_root is not None:
            tree_1 = tree_1.where(Child.parent_id == sa.null() if child_root else
                                  Child.parent_id != sa.null())
        if parent_root is not None:
            tree_1 = tree_1.where(Parent.parent_id == sa.null() if parent_root else
                                  Parent.parent_id != sa.null())

        if node_level is not None:
            if node_level not in LEVEL:  # pragma: no cover
                raise ValueError(f'invalid node_level: {node_level!r}')
            tree_1 = tree_1.where(Node.level == node_level)

        tree_1 = tree_1.cte('tree', recursive=True)

        tree_2 = sa.select(tree_1.c[node_label],
                           recurse_relative.label(relative_label))

        tree_2_onclause = (tree_1.c[relative_label] == join_target)
        if not from_parent:
            tree_2_onclause = sa.and_(tree_2_onclause, recurse_relative != sa.null())

        tree_2_fromclause = tree_1.join(Relative, tree_2_onclause)

        if with_steps:
            tree_2 = tree_2.add_columns((tree_1.c.steps + 1).label('steps'))

        if with_terminal:
            GrandRelative = aliased(cls, name='grand' + ('child'  # noqa: N806
                                                         if from_parent else
                                                         'parent'))
            if from_parent:  # pragma: no cover
                raise NotImplementedError
            tree_2 = tree_2.add_columns((GrandRelative.parent_id == sa.null()).label('terminal'))
            tree_2_fromclause = tree_2_fromclause.outerjoin(GrandRelative,
                                                            Relative.parent_id
                                                            == GrandRelative.id)

        tree_2 = tree_2.select_from(tree_2_fromclause)

        tree = tree_1.union_all(tree_2)

        return tree

    @classmethod
    def tree(cls, /, *, include_self: bool = False,
             with_steps: bool = False,
             with_terminal: bool = False):
        return cls._tree(from_parent=False,
                         innerjoin='reflexive' if include_self else True,
                         with_steps=with_steps, with_terminal=with_terminal)

    @classmethod
    def _path_part(cls, /, *, label: str = 'path_part',
                   include_self: bool = True,
                   bottomup: bool = False,
                   _tree=None):
        if _tree is None:
            tree = cls.tree(include_self=include_self, with_steps=True)
        else:
            tree = _tree

        select_path_part = (sa.select(tree.c.parent_id.label(label))
                            .where(tree.c.child_id == cls.id)
                            .correlate(cls)
                            .order_by(tree.c.steps if bottomup
                                      else tree.c.steps.desc())
                            .alias('parent_path'))

        return select_path_part

    @classmethod
    def path(cls, /, *, label: str = 'path',
             delimiter: str = _globals.FILE_PATH_SEP,
             include_self: bool = True,
             bottomup: bool = False,
             _tree=None):
        squery = cls._path_part(include_self=include_self, bottomup=bottomup,
                                _tree=_tree)
        path = sa.func.group_concat(squery.c.path_part, delimiter).label(label)
        return sa.select(path).label(label)

    @classmethod
    def node_relative(cls, /, *, from_parent: bool = False,
                      innerjoin=False,
                      child_root=None, parent_root=None, node_level=None,
                      with_steps: bool = False,
                      with_terminal: bool = False):
        tree = cls._tree(from_parent=from_parent, innerjoin=innerjoin,
                         child_root=child_root, parent_root=parent_root,
                         node_level=node_level,
                         with_steps=with_steps, with_terminal=with_terminal)

        Child, Parent = cls._aliased_child_parent(child_root=child_root,  # noqa: N806
                                                  parent_root=parent_root)

        if from_parent:
            Node, Relative = Parent, Child  # noqa: N806
            node_label, relative_label = 'parent_id', 'child_id'
        else:
            Node, Relative = Child, Parent  # noqa: N806
            node_label, relative_label = 'child_id', 'parent_id'

        del Child, Parent

        is_node = (tree.c[node_label] == Node.id)
        is_relative = (tree.c[relative_label] == Relative.id)

        node_relative = (tree.join(Node, is_node)
                         .join(Relative, is_relative,
                               isouter=not innerjoin))

        return Node, Relative, tree, node_relative

    @classmethod
    def child_ancestor(cls, /, *, innerjoin=False,
                       child_level=None):
        Child, Parent, _, child_parent = cls.node_relative(from_parent=False,  # noqa: N806
                                                           innerjoin=innerjoin,
                                                           node_level=child_level)
        return Child, Parent, _, child_parent

    @classmethod
    def parent_descendant(cls, /, *, innerjoin=False,
                          parent_root=None, parent_level=None):
        Parent, Child, _, parent_child = cls.node_relative(from_parent=True,  # noqa: N806
                                                           innerjoin=innerjoin,
                                                           parent_root=parent_root,
                                                           node_level=parent_level)
        return Parent, Child, parent_child

    @classmethod
    def path_family_language(cls, /, *, path_label: str = 'path',
                             path_delimiter: str = _globals.FILE_PATH_SEP,
                             include_self: bool = True,
                             bottomup: bool = False,
                             family_label: str = 'family_id',
                             language_label: str = 'language_id'):
        tree = cls.tree(include_self=include_self, with_steps=True, with_terminal=True)

        path = cls.path(label=path_label, delimiter=path_delimiter, bottomup=bottomup, _tree=tree)

        family = (sa.select(tree.c.parent_id)
                  .where(tree.c.child_id == cls.id)
                  .correlate(cls)
                  .where(tree.c.steps > 0)
                  .where(tree.c.terminal == sa.true())
                  .label(family_label))

        Ancestor = aliased(Languoid, name='ancestor')  # noqa: N806

        language = (sa.select(tree.c.parent_id)
                    .where(tree.c.child_id == cls.id)
                    .correlate(cls)
                    .where(cls.level == DIALECT)
                    .where(sa.exists()
                           .where(tree.c.parent_id == Ancestor.id)
                           .where(Ancestor.level == LANGUAGE))
                    .label(language_label))

        return path, family, language


@registry.mapped
class LanguoidLevel:

    __tablename__ = 'languoidlevel'

    name = Column(String, CheckConstraint("name != ''"), primary_key=True)

    description = Column(Text, CheckConstraint("description != ''"),
                         nullable=False)

    ordinal = Column(Integer, CheckConstraint('ordinal >= 1'), nullable=False)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' name={self.name!r}'
                f' description={self.description!r}'
                f' ordinal={self.ordinal}>')

    languoids = relationship('Languoid',
                             back_populates='languoidlevel')


@registry.mapped
class PseudoFamily:

    __tablename__ = 'pseudofamily'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)

    name = Column(ForeignKey('languoid.name'), nullable=False,
                  unique=True)

    config_section = Column(String, CheckConstraint("config_section != ''"),
                            nullable=False, unique=True)

    description = Column(Text, CheckConstraint("description != ''"))

    bookkeeping = Column(sa.Boolean(create_constraint=False), CheckConstraint('bookkeeping = 1'),
                         unique=True)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' name={self.name!r}'
                f' config_section={self.config_section!r}'
                f' description={self.description!r}'
                f' bookkeeping={self.bookkeeping!r}>')

    languoid = relationship('Languoid', foreign_keys=[languoid_id],
                            back_populates='pseudofamily')


@registry.mapped
class Macroarea:

    __tablename__ = 'macroarea'

    name = Column(String, CheckConstraint("name != ''"), primary_key=True)

    config_section = Column(String, CheckConstraint("config_section != ''"),
                            nullable=False, unique=True)

    description = Column(Text, CheckConstraint("description != ''"),
                         nullable=False)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' name={self.name!r}'
                f' config_section={self.config_section!r}'
                f' description={self.description!r}>')

    languoids = relationship('Languoid',
                             secondary='languoid_macroarea',
                             order_by='Languoid.id',
                             back_populates='macroareas')


languoid_macroarea = Table('languoid_macroarea', registry.metadata,
                           Column('languoid_id',
                                  ForeignKey('languoid.id'),
                                  primary_key=True),
                           Column('macroarea_name',
                                  ForeignKey('macroarea.name'),
                                  primary_key=True),
                           info={'without_rowid': True})


@registry.mapped
class Country:

    __tablename__ = 'country'

    id = Column(String(2), CheckConstraint('length(id) = 2'), primary_key=True)

    name = Column(Text, CheckConstraint("name != ''"), nullable=False,
                  unique=True)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' id={self.id!r}'
                f' name={self.name!r}>')

    languoids = relationship('Languoid',
                             secondary='languoid_country',
                             order_by='Languoid.id',
                             back_populates='countries')

    @classmethod
    def printf(cls, /, *, minimal: bool = True, label: str = 'printf'):
        return (sa.func.printf('%s (%s)', cls.name, cls.id)
                if not minimal else cls.id).label(label)

    @classmethod
    def jsonf(cls, /, *, sort_keys: bool, label: str = 'jsonf'):
        return json_object(id=cls.id,
                           name=cls.name,
                           sort_keys_=sort_keys,
                           label_=label,)


languoid_country = Table('languoid_country', registry.metadata,
                         Column('languoid_id',
                                ForeignKey('languoid.id'),
                                primary_key=True),
                         Column('country_id',
                                ForeignKey('country.id'),
                                primary_key=True),
                         info={'without_rowid': True})


@registry.mapped
class Link:

    __tablename__ = 'link'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    ord = Column(Integer, CheckConstraint('ord >= 1'), primary_key=True)

    url = Column(Text, CheckConstraint("url != ''"), nullable=False)

    title = Column(Text, CheckConstraint("title != ''"))
    scheme = Column(Text, Enum(*sorted(LINK_SCHEME), create_constraint=True))

    __table_args__ = (UniqueConstraint(languoid_id, url),
                      CheckConstraint("substr(url, 1, length(scheme) + 3)"
                                      " = scheme || '://'"),
                      {'info': {'without_rowid': True}})

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' ord={self.ord!r}'
                f' url={self.url!r}'
                f' title={self.title!r}'
                f' scheme={self.scheme!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='links')

    @classmethod
    def printf(cls, /, *, label: str = 'printf'):
        return sa.case((cls.title != sa.null(),
                        sa.func.printf('[%s](%s)', cls.title, cls.url)),
                       else_=cls.url).label(label)

    @classmethod
    def jsonf(cls, /, *, sort_keys: bool, label: str = 'jsonf'):
        return json_object(scheme=cls.scheme,
                           url=cls.url,
                           title=cls.title,
                           sort_keys_=sort_keys,
                           label_=label)


@registry.mapped
class Timespan:

    __tablename__ = 'timespan'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)

    start_year = Column(Integer,
                        CheckConstraint('start_year BETWEEN -999999 AND 999999'),
                        nullable=False)
    start_month = Column(Integer,
                         CheckConstraint('start_month BETWEEN 1 AND 12'),
                         nullable=False)
    start_day = Column(Integer,
                       CheckConstraint('start_day BETWEEN 1 AND 31'),
                       nullable=False)

    end_year = Column(Integer,
                      CheckConstraint('end_year BETWEEN -999999 AND 999999'),
                      nullable=False)
    end_month = Column(Integer,
                       CheckConstraint('end_month BETWEEN 1 AND 12'),
                       nullable=False)
    end_day = Column(Integer,
                     CheckConstraint('end_day BETWEEN 1 AND 31'),
                     nullable=False)

    __table_args__ = (CheckConstraint('end_year - start_year >= 0'),
                      {'info': {'without_rowid': True}})

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' start_year={self.start_year!r}'
                f' start_month={self.start_month!r}'
                f' start_day={self.start_day!r}'
                f' end_year={self.end_year!r}'
                f' end_month={self.end_month!r}'
                f' end_day={self.end_day!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='timespan')

    @classmethod
    def printf(cls, /, *, label: str = 'printf'):
        return sa.func.printf('%s-%s-%s/%s-%s-%s',
                              cls.start_year, cls.start_month, cls.start_day,
                              cls.end_year, cls.end_month, cls.end_day)

    @classmethod
    def jsonf(cls, /, *, sort_keys: bool, label: str = 'jsonf'):
        return json_object(start_year=cls.start_year,
                           start_month=cls.start_month,
                           start_day=cls.start_day,
                           end_year=cls.end_year,
                           end_month=cls.end_month,
                           end_day=cls.end_day,
                           sort_keys_=sort_keys,
                           label_=label)


@registry.mapped
class Source:

    __tablename__ = 'source'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    provider_id = Column(ForeignKey('sourceprovider.id'), primary_key=True)
    bibitem_id = Column(ForeignKey('bibitem.id'), primary_key=True)

    pages = Column(Text, CheckConstraint("pages != ''"))
    trigger = Column(Text, CheckConstraint("trigger != ''"))

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' provider_id={self.provider_id!r}'
                f' bibitem_id={self.bibitem_id!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='sources')

    provider = relationship('SourceProvider',
                            innerjoin=True,
                            back_populates='sources')

    bibitem = relationship('Bibitem',
                           innerjoin=True,
                           back_populates='sources')

    @classmethod
    def printf(cls, bibfile, bibitem, /, *,
               label: str = 'printf'):
        return sa.case((sa.and_(cls.pages != sa.null(), cls.trigger != sa.null()),
                        sa.func.printf('**%s:%s**:%s<trigger "%s">',
                                       bibfile.name, bibitem.bibkey,
                                       cls.pages, cls.trigger)),
                       (cls.pages != sa.null(),
                        sa.func.printf('**%s:%s**:%s',
                                       bibfile.name, bibitem.bibkey,
                                       cls.pages)),
                       (cls.trigger != sa.null(),
                        sa.func.printf('**%s:%s**<trigger "%s">',
                                       bibfile.name, bibitem.bibkey,
                                       cls.trigger)),
                       else_=sa.func.printf('**%s:%s**',
                                            bibfile.name, bibitem.bibkey)
                       ).label(label)

    @classmethod
    def jsonf(cls, bibfile, bibitem, /, *,
              sort_keys: bool, label: str = 'jsonf'):
        return json_object(bibfile=bibfile.name,
                           bibkey=bibitem.bibkey,
                           pages=cls.pages,
                           trigger=cls.trigger,
                           sort_keys_=sort_keys,
                           label_=label)


@registry.mapped
class SourceProvider:

    __tablename__ = 'sourceprovider'

    id = Column(Integer, primary_key=True)

    name = Column(Text, Enum(*sorted(SOURCE_PROVIDER), create_constraint=True),
                  nullable=False, unique=True)

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' name={self.name!r}>')

    sources = relationship('Source', back_populates='provider')


@registry.mapped
class Bibfile:

    __tablename__ = 'bibfile'

    id = Column(Integer, primary_key=True)

    name = Column(String, CheckConstraint("name != ''"), nullable=False,
                  unique=True)

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' id={self.id!r}'
                f' name={self.name!r}>')

    bibitems = relationship('Bibitem', back_populates='bibfile')


@registry.mapped
class Bibitem:

    __tablename__ = 'bibitem'

    id = Column(Integer, primary_key=True)

    bibfile_id = Column(ForeignKey('bibfile.id'), nullable=False)
    bibkey = Column(Text, CheckConstraint("bibkey != ''"), nullable=False)

    __table_args__ = (UniqueConstraint(bibfile_id, bibkey),)

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' bibfile_id={self.bibfile_id!r}'
                f' bibkey={self.bibkey!r}>')

    bibfile = relationship('Bibfile',
                           innerjoin=True,
                           back_populates='bibitems')

    sources = relationship('Source', back_populates='bibitem')

    classificationrefs = relationship('ClassificationRef',
                                      back_populates='bibitem')

    endangermentstatus = relationship('EndangermentStatus',
                                      back_populates='bibitem')

    endangermentsources = relationship('EndangermentSource',
                                       back_populates='bibitem')


@registry.mapped
class Altname:

    __tablename__ = 'altname'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    provider_id = Column(ForeignKey('altnameprovider.id'), primary_key=True)
    name = Column(Text, CheckConstraint("name != ''"), primary_key=True)
    lang = Column(String(3), CheckConstraint('length(lang) IN (0, 2, 3)'),
                  server_default='', primary_key=True)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' provider_id={self.provider_id!r}'
                f' name={self.name!r}'
                f' lang={self.lang!r}>')

    languoid = relationship('Languoid', innerjoin=True,
                            back_populates='altnames')

    provider = relationship('AltnameProvider',
                            innerjoin=True,
                            back_populates='altnames')

    @classmethod
    def printf(cls, /, *, label: str = 'printf'):
        full = sa.func.printf('%s [%s]', cls.name, cls.lang)
        return sa.case((cls.lang == '', cls.name), else_=full).label(label)

    @classmethod
    def jsonf(cls, /, *, sort_keys: bool, label: str = 'jsonf'):
        half = json_object(name=cls.name,
                           lang=None,
                           sort_keys_=sort_keys)
        full = json_object(name=cls.name,
                           lang=cls.lang,
                           sort_keys_=sort_keys)
        return sa.case((cls.lang == '', half), else_=full).label(label)


@registry.mapped
class AltnameProvider:

    __tablename__ = 'altnameprovider'

    id = Column(Integer, primary_key=True)

    name = Column(Text, Enum(*sorted(ALTNAME_PROVIDER), create_constraint=True),
                  nullable=False, unique=True)

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' name={self.name!r}>')

    altnames = relationship('Altname', back_populates='provider')


@registry.mapped
class Trigger:

    __tablename__ = 'trigger'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    field = Column(Enum(*sorted(TRIGGER_FIELD), create_constraint=True),
                   primary_key=True)
    trigger = Column(Text, CheckConstraint("trigger != ''"), primary_key=True)

    ord = Column(Integer, CheckConstraint('ord >= 1'), nullable=False)

    __table_args__ = (UniqueConstraint(languoid_id, field, ord),
                      {'info': {'without_rowid': True}})

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' field={self.field!r}'
                f' trigger={self.trigger!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='triggers')


@registry.mapped
class Identifier:

    __tablename__ = 'identifier'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    site_id = Column(ForeignKey('identifiersite.id'), primary_key=True)

    identifier = Column(Text, CheckConstraint("identifier != ''"), nullable=False)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' site_id={self.site_id!r}'
                f' identifier={self.identifier!r}>')

    languoid = relationship('Languoid', innerjoin=True,
                            back_populates='identifiers')

    site = relationship('IdentifierSite', innerjoin=True,
                        back_populates='identifiers')


@registry.mapped
class IdentifierSite:

    __tablename__ = 'identifiersite'

    id = Column(Integer, primary_key=True)

    name = Column(Text, Enum(*sorted(IDENTIFIER_SITE), create_constraint=True),
                  nullable=False, unique=True)

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' name={self.name!r}>')

    identifiers = relationship('Identifier', back_populates='site')


@registry.mapped
class ClassificationComment:

    __tablename__ = 'classificationcomment'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    kind = Column(Enum(*sorted(CLASSIFICATION_KIND), create_constraint=True),
                  primary_key=True)

    comment = Column(Text, CheckConstraint("comment != ''"), nullable=False)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' kind={self.kind!r}'
                f' comment={self.comment!r}>')

    languoid = relationship('Languoid', innerjoin=True)


@registry.mapped
class ClassificationRef:

    __tablename__ = 'classificationref'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)
    kind = Column(Enum(*sorted(CLASSIFICATION_KIND), create_constraint=True),
                  primary_key=True)
    bibitem_id = Column(ForeignKey('bibitem.id'), primary_key=True)

    ord = Column(Integer, CheckConstraint('ord >= 1'), nullable=False)

    pages = Column(Text, CheckConstraint("pages != ''"))

    __table_args__ = (UniqueConstraint(languoid_id, kind, ord),
                      {'info': {'without_rowid': True}})

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' kind={self.kind!r}'
                f' bibitem_id={self.bibitem_id!r}>')

    languoid = relationship('Languoid', innerjoin=True)

    bibitem = relationship('Bibitem',
                           innerjoin=True,
                           back_populates='classificationrefs')

    @classmethod
    def printf(cls, bibfile, bibitem, /, *,
               label: str = 'printf'):
        return sa.func.printf(sa.case((cls.pages != sa.null(), '**%s:%s**:%s'),
                                      else_='**%s:%s**'),
                              bibfile.name, bibitem.bibkey, cls.pages).label(label)

    @classmethod
    def jsonf(cls, bibfile, bibitem, /, *,
              sort_keys: bool, label: str = 'jsonf'):
        return json_object(bibfile=bibfile.name,
                           bibkey=bibitem.bibkey,
                           pages=cls.pages,
                           trigger=None,
                           sort_keys_=sort_keys,
                           label_=label)


@registry.mapped
class Endangerment:

    __tablename__ = 'endangerment'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)

    status = Column(ForeignKey('endangermentstatus.name'), nullable=False)

    source_id = Column(ForeignKey('endangermentsource.id'), nullable=False)
    date = Column(DateTime, nullable=False)
    comment = Column(Text, CheckConstraint("comment != ''"), nullable=False)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' status={self.status!r}'
                f' source_id={self.source_id!r}'
                f' date={self.date!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='endangerment')

    endangermentstatus = relationship('EndangermentStatus',
                                      innerjoin=True,
                                      back_populates='endangerments')

    source = relationship('EndangermentSource',
                          innerjoin=True,
                          back_populates='endangerment')

    @classmethod
    def jsonf(cls, source, bibfile, bibitem, /, *,
              sort_keys: bool, label: str = 'jsonf'):
        source = json_object(name=source.name,
                             bibfile=bibfile.name,
                             bibkey=bibitem.bibkey,
                             pages=source.pages,
                             sort_keys_=sort_keys)
        return json_object(status=cls.status,
                           source=source,
                           date=json_datetime(cls.date),
                           comment=cls.comment,
                           sort_keys_=sort_keys,
                           label_=label)


@registry.mapped
class EndangermentStatus:

    __tablename__ = 'endangermentstatus'

    name = Column(String, CheckConstraint("name != ''"), primary_key=True)

    config_section = Column(String, CheckConstraint("config_section != ''"),
                            nullable=False, unique=True)

    ordinal = Column(Integer, CheckConstraint('ordinal >= 1'), nullable=False)

    egids = Column(String, CheckConstraint("egids != ''"), nullable=False)
    unesco = Column(String, CheckConstraint("unesco != ''"), nullable=False)
    elcat = Column(String, CheckConstraint("elcat != ''"), nullable=False)
    icon = Column(String, CheckConstraint("icon != ''"), nullable=False)

    bibitem_id = Column(ForeignKey('bibitem.id'))

    __table_args__ = {'info': {'without_rowid': True}}

    bibitem = relationship('Bibitem',
                           back_populates='endangermentstatus')

    endangerments = relationship('Endangerment',
                                 back_populates='endangermentstatus')


@registry.mapped
class EndangermentSource:

    __tablename__ = 'endangermentsource'

    id = Column(Integer, primary_key=True)
    name = Column(Text, CheckConstraint("name != ''"), nullable=False, unique=True)

    bibitem_id = Column(ForeignKey('bibitem.id'))
    pages = Column(Text, CheckConstraint("pages != ''"))

    __table_args__ = (UniqueConstraint(bibitem_id, pages),
                      CheckConstraint('(bibitem_id IS NULL) = (pages IS NULL)'))

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' id={self.id!r}'
                f' name={self.name!r}'
                f' bibitem_id={self.bibitem_id!r}'
                f' pages={self.pages!r}>')

    bibitem = relationship('Bibitem',
                           back_populates='endangermentsources')

    endangerment = relationship('Endangerment',
                                uselist=False,
                                back_populates='source')

    @classmethod
    def printf(cls, bibfile, bibitem, /, *,
               label: str = 'printf'):
        return sa.case((cls.bibitem_id == sa.null(), cls.name),
                       else_=sa.func.printf('**%s:%s**:%s',
                                            bibfile.name, bibitem.bibkey,
                                            cls.pages)).label(label)


@registry.mapped
class EthnologueComment:

    __tablename__ = 'ethnologuecomment'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)

    isohid = Column(Text, CheckConstraint('length(isohid) >= 3'), nullable=False)
    comment_type = Column(Enum(*sorted(EL_COMMENT_TYPE), create_constraint=True),
                          nullable=False)
    ethnologue_versions = Column(Text, CheckConstraint('length(ethnologue_versions) >= 3'), nullable=False)
    comment = Column(Text, CheckConstraint("comment != ''"), nullable=False)

    __table_args__ = {'info': {'without_rowid': True}}

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' isohid={self.isohid!r}'
                f' comment_type={self.comment_type!r}'
                f' ethnologue_versions={self.ethnologue_versions!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='ethnologue_comment')

    @classmethod
    def jsonf(cls, /, *, sort_keys: bool, optional: bool = False,
              label: str = 'jsonf'):
        mapping = json_object(isohid=cls.isohid,
                              comment_type=cls.comment_type,
                              ethnologue_versions=cls.ethnologue_versions,
                              comment=cls.comment,
                              sort_keys_=sort_keys)
        if optional:
            return sa.case((cls.languoid_id == sa.null(), None), else_=mapping).label(label)
        return mapping.label(label)


@registry.mapped
class IsoRetirement:

    __tablename__ = 'isoretirement'

    languoid_id = Column(ForeignKey('languoid.id'), primary_key=True)

    code = Column(String(3), CheckConstraint('length(code) = 3'), nullable=False)
    name = Column(Text, CheckConstraint("name != ''"), nullable=False)

    change_request = Column(String(8), CheckConstraint("change_request LIKE '____-___' "))

    effective = Column(Date, nullable=False)
    reason = Column(Enum(*sorted(ISORETIREMENT_REASON), create_constraint=True),
                    nullable=False)

    remedy = Column(Text, CheckConstraint("remedy != ''"))
    comment = Column(Text, CheckConstraint("comment != ''"))

    __table_args__ = (
        # TODO: fix disagreement
        Index('change_request_key', sa.func.coalesce(change_request, effective)),
        CheckConstraint("reason != 'split' or remedy IS NOT NULL"),
        {'info': {'without_rowid': True}},
    )

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' code={self.code!r}'
                f' name={self.name!r}'
                f' change_request={self.change_request!r}'
                f' effective={self.effective!r}'
                f' reason={self.reason!r}'
                f' remedy={self.remedy!r}>')

    languoid = relationship('Languoid',
                            innerjoin=True,
                            back_populates='iso_retirement')

    change_to = relationship('IsoRetirementChangeTo',
                             order_by='IsoRetirementChangeTo.ord',
                             back_populates='iso_retirement')

    @classmethod
    def jsonf(cls, /, *, sort_keys: bool,
              change_to=None, optional: bool = False,
              label: str = 'jsonf'):
        mapping = json_object(code=cls.code,
                              name=cls.name,
                              change_request=cls.change_request,
                              change_to=change_to,
                              effective=json_datetime(cls.effective),
                              reason=cls.reason,
                              remedy=cls.remedy,
                              comment=cls.comment,
                              sort_keys_=sort_keys)
        if optional:
            return sa.case((cls.languoid_id == sa.null(), None), else_=mapping).label(label)
        return mapping.label(label)


@registry.mapped
class IsoRetirementChangeTo:

    __tablename__ = 'isoretirement_changeto'

    languoid_id = Column(ForeignKey('isoretirement.languoid_id'), primary_key=True)
    code = Column(String(3), CheckConstraint('length(code) = 3'), primary_key=True)

    ord = Column(Integer, CheckConstraint('ord >= 1'), nullable=False)

    __table_args__ = (UniqueConstraint('languoid_id', 'ord'),
                      {'info': {'without_rowid': True}})

    def __repr__(self):
        return (f'<{self.__class__.__name__}'
                f' languoid_id={self.languoid_id!r}'
                f' code={self.code!r}>')

    iso_retirement = relationship('IsoRetirement',
                                  innerjoin=True,
                                  back_populates='change_to')



"""Load, parse, serialize, and write ``languoids/tree/**/md.ini`` files."""

from collections.abc import Iterable

from .. import _globals
from .. import _tools

from ._root import TREE_IN_ROOT, set_root, get_repo_root
from .files import iterfiles

__all__ = ['TREE_IN_ROOT', 'set_root', 'get_repo_root',
           'iterfiles',
           'iterrecords']


def iterrecords(*, root=_globals.ROOT,
                progress_after: int = _tools.PROGRESS_AFTER
                ) -> Iterable[_globals.RecordItem]:
    for path_tuple, _, cfg in iterfiles(root, progress_after=progress_after):
        yield path_tuple, cfg



"""SQLite3 database export functions."""

import contextlib
import datetime
import functools
import gzip
import hashlib
import logging
import pprint
import warnings
import zipfile

import csv23

import sqlalchemy as sa

from .. import _globals
from .. import _tools
from .. import backend as _backend

from .models import Dataset, Producer

__all__ = ['print_dataset',
           'print_schema',
           'print_query_sql',
           'get_query_sql',
           'backup',
           'dump_sql',
           'csv_zipfile',
           'print_rows',
           'write_csv', 'hash_csv', 'hash_rows']


log = logging.getLogger(__name__)


def print_dataset(*, ignore_dirty: bool = False,
                  file=None,
                  bind=_globals.ENGINE):
    with _backend.connect(bind=bind) as conn:
        dataset = (conn.execute(sa.select(Dataset))
                   .mappings()
                   .one())
        producer = (conn.execute(sa.select(Producer))
                    .mappings()
                    .one())
    Dataset.log_dataset(dataset, ignore_dirty=ignore_dirty,
                        also_print=True, print_file=file)
    Producer.log_producer(producer,
                          also_print=True, print_file=file)


def print_schema(metadata=_globals.REGISTRY.metadata, /, *,
                 file=None,
                 engine=_globals.ENGINE):
    """Print the SQL from metadata.create_all() without executing."""
    def print_sql(sql, *_, **__):
        print(sql.compile(dialect=engine.dialect),
              file=file)

    mock_engine = sa.create_mock_engine(engine.url, executor=print_sql)

    metadata.create_all(mock_engine, checkfirst=False)


def print_query_sql(query=None, /, *, literal_binds: bool = True,
                    pretty: bool = True,
                    file=None, flush: bool = True):
    """Print the literal SQL for the given query."""
    sql = get_query_sql(query, literal_binds=literal_binds, pretty=pretty)
    print(sql, file=file, flush=flush)


def get_query_sql(query=None, /, *, literal_binds: bool = True,
                  pretty: bool = False):
    """Return the literal SQL for the given query."""
    if query is None:
        from .. import queries

        query = queries.get_example_query()

    compiled = _backend.expression_compile(query, literal_binds=literal_binds)
    result = compiled.string

    if pretty and _backend.sqlparse is not None:
        result = _backend.sqlparse.format(result, reindent=True)
    return result


def backup(filename=None, /, *, as_new_engine: bool = False,
           pages: int = 0,
           engine=_globals.ENGINE):
    """Write the database into another .sqlite3 file and return its engine."""
    log.info('backup database')
    log.info('source: %r', engine)

    url = 'sqlite://'
    if filename is not None:
        path = _tools.path_from_filename(filename)
        if path.exists():
            if engine.file is not None and path.samefile(engine.file):
                raise ValueError(f'backup destination {path!r} same file as'
                                 f' source {engine.file!r}')
            warnings.warn(f'delete present file: {path!r}')
            path.unlink()
        url += f'/{path}'

    log.info('destination: %r', url)
    result = sa.create_engine(url, future=engine.future)

    def progress(status, remaining, total):
        log.info('%d of %d pages copied', total - remaining, total)

    with (contextlib.closing(engine.raw_connection()) as source_fairy,
          contextlib.closing(result.raw_connection()) as dest_fairy):
        log.debug('sqlite3.backup(%r)', dest_fairy.driver_connection)

        dest_fairy.execute('PRAGMA synchronous = OFF')
        dest_fairy.execute('PRAGMA journal_mode = MEMORY')

        with dest_fairy.driver_connection as dbapi_conn:
            source_fairy.backup(dbapi_conn, pages=pages, progress=progress)

    log.info('database backup complete')
    if as_new_engine:
        _backend.set_engine(result)
    return result


def dump_sql(filename=None, /, *,
             progress_after: int = 100_000,
             encoding: str = _tools.ENCODING,
             engine=_globals.ENGINE):
    """Dump the engine database into a plain-text SQL file."""
    if filename is None:
        filename = engine.file_with_suffix('.sql.gz').name
    path = _tools.path_from_filename(filename)
    log.info('dump sql to %r', path)

    if path.exists():
        warnings.warn(f'delete present file: {path!r}')
        path.unlink()

    if path.suffix == '.gz':
        open_path = functools.partial(gzip.open, path)
    else:
        open_path = path.open

    n = 0
    with (contextlib.closing(engine.raw_connection()) as dbapi_fairy,
          open_path('wt', encoding=encoding) as f):
        for n, line in enumerate(dbapi_fairy.iterdump(), start=1):
            print(line, file=f)
            if not (n % progress_after):
                log.info('%s lines written', f'{n:_d}')

    log.info('%s lines total', f'{n:_d}')
    return path


def csv_zipfile(filename=None, /, *, exclude_raw: bool = False,
                metadata=_globals.REGISTRY.metadata,
                dialect=csv23.DIALECT, encoding: str = csv23.ENCODING,
                engine=_globals.ENGINE):
    """Write all tables to <tablename>.csv in <databasename>.zip."""
    log.info('export database')
    log.debug('engine: %r', engine)

    if filename is None:
        filename = engine.file_with_suffix('.zip').name

    path = _tools.path_from_filename(filename)
    if path.exists():
        warnings.warn(f'delete present file: {path!r}')
        path.unlink()

    filename = str(path)

    sorted_tables = sorted(metadata.sorted_tables, key=lambda t: t.name)

    skip = {'_file', '_option', '_value'} if exclude_raw else {}

    log.info('write %r', filename)
    with (_backend.connect(bind=engine) as conn,
          zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as z):
        for table in sorted_tables:
            if table.name in skip:
                log.debug('skip table %r', table.name)
                continue

            log.info('export table %r', table.name)
            rows = conn.execute(sa.select(table))
            header = list(rows.keys())

            date_time = datetime.datetime.now().timetuple()[:6]
            info = zipfile.ZipInfo(f'{table.name}.csv', date_time=date_time)
            info.compress_type = zipfile.ZIP_DEFLATED

            with z.open(info, 'w') as f:
                csv23.write_csv(f, rows, header=header,
                                dialect=dialect, encoding=encoding)

    log.info('database export complete.')
    return _tools.path_from_filename(filename)


def print_rows(query=None, /, *, file=None,
               pretty: bool = False,
               format_: str | None = None,
               verbose: bool = False,
               mappings: bool = True,
               bind=_globals.ENGINE):
    if query is None:
        from .. import queries as _queries

        query = _queries.get_example_query()

    if not isinstance(query, sa.sql.base.Executable):
        # assume mappings
        rows = iter(query)
    else:
        if verbose:
            print(query, file=file)

        rows = _backend.iterrows(query,
                                 mappings=(format_ is not None) or mappings,
                                 bind=bind)

    print_func = functools.partial(print, file=file)
    if format_ is not None:
        rows = map(format_.format_map, rows)
    elif pretty:
        rows = map(dict if mappings else tuple, rows)
        print_func = PrettyPrinter(file).pprint

    for r in rows:
        print_func(r)


class PrettyPrinter(pprint.PrettyPrinter):

    def __init__(self, stream, /, **kwargs) -> None:
        kwargs.setdefault('sort_dicts', False)
        kwargs.setdefault('underscore_numbers', True)
        super().__init__(stream=stream, **kwargs)


def write_csv(query=None, /, filename=None, *,
              verbose: bool = False,
              dialect=csv23.DIALECT, encoding: str = csv23.ENCODING,
              bind=_globals.ENGINE):
    """Write get__example_query() query (or given query) to CSV, return filename."""
    if query is None:
        from .. import queries as _queries

        query = _queries.get_example_query()

    if filename is None:
        filename = bind.file_with_suffix('.query.csv').name
    filename = _tools.path_from_filename(filename)

    log.info('write csv: %r', filename)
    path = _tools.path_from_filename(filename)
    if path.exists():
        warnings.warn(f'delete present file: {path!r}')
        path.unlink()

    if verbose:
        print(query)

    with _backend.connect(bind=bind) as conn:
        result = conn.execute(query)

        header = list(result.keys())
        log.info('csv header: %r', header)
        return csv23.write_csv(filename, result, header=header,
                               dialect=dialect, encoding=encoding,
                               autocompress=True)


def hash_csv(query=None, /, *, hash_name: str = _globals.DEFAULT_HASH,
             dialect=csv23.DIALECT, encoding: str = csv23.ENCODING,
             raw: bool = False,
             bind=_globals.ENGINE):
    if query is None:
        from .. import queries as _queries

        query = _queries.get_example_query()

    with _backend.connect(bind=bind) as conn:
        result = conn.execute(query)

        header = list(result.keys())
        return hash_rows(result, header=header, hash_name=hash_name, raw=raw,
                         dialect=dialect, encoding=encoding)


def hash_rows(rows, /, *, hash_name: str = _globals.DEFAULT_HASH,
              header=None,
              dialect=csv23.DIALECT, encoding=csv23.ENCODING,
              raw: bool = False):
    if hash_name is None:
        hash_name = _globals.DEFAULT_HASH

    log.info('hash rows with %r, csv header: %r', hash_name, header)
    result = hashlib.new(hash_name)
    assert hasattr(result, 'hexdigest')

    csv23.write_csv(result, rows, header=header,
                    dialect=dialect, encoding=encoding)

    if not raw:
        result = result.hexdigest()
    return result



"""Import data into SQLite3 database."""

import datetime
import functools
import logging
import time
import warnings

import sqlalchemy as sa

from .. import _globals
from .. import _tools
from .. import backend as _backend
from .. import languoids as _languoids
from .. import config as _config
from .. import glottolog as _glottolog

from . import models as _models
from . import views as _views

__all__ = ['main']


log = logging.getLogger(__name__)


def get_root(repo_root, /, *, default,
             treepath=_languoids.TREE_IN_ROOT):
    if repo_root is not None:
        root = _languoids.set_root(repo_root, treepath=treepath)
    else:
        root = default

    log.info('load database from %r', root)
    if not root.exists():
        log.error('root does not exist')
        raise RuntimeError(f'tree root not found: {root!r}')
    return root


def get_from_raw(from_raw, /, *, exclude_raw: bool):
    if exclude_raw and from_raw:  # pragma: no cover
        log.error('incompatible exclude_raw=%r and from_raw=%r', exclude_raw, from_raw)
        raise ValueError('exclude_raw and from_raw cannot both be True')
    elif from_raw is None:
        from_raw = not exclude_raw
    return from_raw


def get_engine(filename_or_engine, /, *, require: bool):
    if hasattr(filename_or_engine, 'execute'):
        return filename_or_engine
    return _backend.set_engine(filename_or_engine, require=require)


def get_dataset(engine, /, *, exclude_raw: bool, strict: bool):
    dataset = None

    if engine.file is None:
        log.warning('connected to a transient in-memory database')
        dataset = _models.Dataset.get_dataset(bind=engine, strict=True)
    elif engine.file_size():
        dataset = _models.Dataset.get_dataset(bind=engine, strict=strict)

        if dataset is None:
            warnings.warn(f'force delete {engine.file!r}')
        elif dataset.exclude_raw != bool(exclude_raw):  # pragma: no cover
            dataset = None
            log.warning('rebuild needed from exclude_raw mismatch')

    return dataset


def main(filename=_globals.ENGINE, repo_root=None, /, *,
         treepath=_languoids.TREE_IN_ROOT,
         metadata=_globals.REGISTRY.metadata,
         require: bool = False,
         rebuild: bool = False,
         from_raw: bool = None,
         exclude_raw: bool = False,
         exclude_views: bool = False,
         force_rebuild: bool = False,
         _only_create_tables: bool = False):
    """Load languoids/tree/**/md.ini into SQLite3 db, return engine."""
    kwargs = {'root': get_root(repo_root, default=_globals.ROOT, treepath=treepath),
              'from_raw': get_from_raw(from_raw, exclude_raw=exclude_raw)}

    engine = get_engine(filename, require=require)

    dataset = get_dataset(engine,
                          exclude_raw=exclude_raw,
                          strict=not force_rebuild and not _only_create_tables)

    if dataset is None or rebuild or _only_create_tables:
        log.info('build new database' if dataset is None else 'rebuild database')
        engine.dispose()
        if engine.file_size():
            warnings.warn(f'delete present file: {engine.file!r}')
            engine.file.unlink()

        log.info('create %d tables from %r', len(metadata.tables), metadata)
        with _backend.connect(bind=engine) as conn:
            create_tables(metadata, conn=conn,
                          exclude_raw=exclude_raw,
                          exclude_views=exclude_views)
            log.info('COMMIT create_all: %r', conn)
            conn.commit()

        if _only_create_tables:
            return engine

        log.debug('start load timer')
        start = time.time()
        with _backend.connect(bind=engine, pragma_bulk_insert=True) as conn:
            dataset = load(metadata, conn=conn,
                           exclude_raw=exclude_raw,
                           **kwargs)
        walltime = datetime.timedelta(seconds=time.time() - start)
        log.debug('load timer stopped')
        print(walltime)

        log.info('database load complete.')
    else:
        log.info('use present %r', engine)

    _models.Dataset.log_dataset(dataset)
    pdc = _models.Producer.get_producer(bind=engine)
    _models.Producer.log_producer(pdc)

    return engine


def create_tables(metadata, /, *, conn,
                  exclude_raw: bool, exclude_views: bool):
    # import here to register models for create_all()
    log.debug('import module %s.models', __package__)
    from .. import models

    assert models is not None

    if not exclude_raw:
        log.debug('import module %s.raw', __package__)

        from .. import raw

        assert raw is not None

    if not exclude_views:
        log.info('prepare %d views', len(_views.REGISTERED))

    try:
        _views.create_all_views(clear=exclude_views)
    except Exception:  # pragma: no cover
        log.exception('error running %s.views.create_all_views(clear=%r)',
                      __package__, exclude_views)

    application_id = sum(ord(c) for c in _models.Dataset.__tablename__)
    assert application_id == 1122 == 0x462

    log.debug('set application_id = %r', application_id)
    conn.execute(sa.text(f'PRAGMA application_id = {application_id:d}'))

    log.debug('run create_all')
    metadata.create_all(bind=conn)


def load(metadata, /, *, conn, root,
         from_raw: bool, exclude_raw: bool):
    log.info('record git commit in %r', root)
    # pre-create dataset to added as final item marking completeness
    dataset = make_dataset(root, exclude_raw=exclude_raw)
    _models.Dataset.log_dataset(dataset)

    log.info('write %r', _models.Producer.__tablename__)
    write_producer(conn, name=__package__.partition('.')[0])

    log.info('load config/*.ini into %r', _models.Config.__tablename__)
    version = import_configs(conn, root=root)
    log.info('version from %r: %r', _models.Config.__tablename__, version)
    dataset['version'] = version

    log.info('COMMIT producer and configs: %r', conn)
    conn.commit()

    if not exclude_raw:
        log.info('load raw')
        import_raw(conn, root=root)

        log.info('COMMIT raw: %r', conn)
        conn.commit()

    if not (from_raw or exclude_raw):  # pragma: no cover
        warnings.warn('2 tree reads required (use compare_with_files() to verify)')

    log.info('load languoids')
    import_languoids(conn, root=root,
                     source='raw' if from_raw else 'files')

    log.info('COMMIT languoids: %r', conn)
    conn.commit()

    log.info('write %r: %r', _models.Dataset.__tablename__, dataset['title'])
    write_dataset(conn, dataset=dataset)

    log.info('COMMIT dataset: %r', conn)
    conn.commit()

    return dataset


def import_configs(conn, /, *, root):
    insert_config = functools.partial(conn.execute, sa.insert(_models.Config))
    for filename, cfg in _config.iterconfigs(root):
        get_line = _tools.next_count(start=1)
        params = [{'filename': filename, 'section': section, 'option': option,
                   'value': value.strip(), 'line': get_line()}
                  for section, sec in cfg.items()
                  for option, value in sec.items()
                  if value.strip()]

        log.debug('insert %d values for %r', len(params), filename)
        insert_config(params)

    select_version = (sa.select(_models.Config.value)
                      .filter_by(filename='publication.ini',
                                 section='zenodo', option='version'))
    return conn.execute(select_version).scalar_one_or_none()


def make_dataset(root, /, *, exclude_raw: bool):
    try:
        dataset = {'title': 'Glottolog treedb',
                   'git_commit': _glottolog.git_rev_parse(root),
                   'git_describe': _glottolog.git_describe(root),
                   # clean = neither changes in index nor untracked files
                   'clean': _glottolog.git_status_is_clean(root),
                   'exclude_raw': exclude_raw}
    except Exception as e:  # pragma: no cover
        log.exception('error running git command in %r', str(root))
        raise RuntimeError(f'failed to get info for dataset: {e}') from e
    else:
        log.info('identified dataset')
        return dataset


def write_dataset(conn, /, *, dataset):
    log.debug('dataset: %r', dataset)
    conn.execute(sa.insert(_models.Dataset), dataset)


def write_producer(conn, /, *, name: str):
    from .. import __version__

    params = {'name': name, 'version': __version__}
    _models.Producer.log_producer(params)
    conn.execute(sa.insert(_models.Producer), params)


def import_raw(conn, /, *, root):
    log.debug('import target module %s.raw.import_models', __package__)

    from ..raw import import_models

    log.debug('root: %r', root)

    import_models.main(root, conn=conn)


def import_languoids(conn, /, *, root, source: str):
    log.debug('import source module %s.languoids', __package__)

    from .. import export

    log.debug('import target module %s.import_models', __package__)

    from .. import import_models

    if source == 'files':
        order_by = True
        root_or_bind = root
    elif source == 'raw':
        # insert languoids in Glottocode order when reading from raw
        order_by = 'id'
        root_or_bind = conn
    else:  # pragma: no cover
        ValueError(f'unknown source: {source!r}')
    log.debug('root_or_bind: %r', root_or_bind)

    pairs = export.iterlanguoids(source,
                                 order_by=order_by,
                                 root=root, bind=conn)

    import_models.main(pairs, conn=conn)



"""Define known ``md.ini`` (section, option) pairs and whether they are lists of lines."""

from collections.abc import Mapping
import logging
from typing import TypeAlias
import warnings

from .. import _globals
from .. import _tools

__all__ = ['is_lines',
           'sorted_sections',
           'sorted_options',
           'parse_lines', 'format_lines',
           'join_lines_inplace']

SECTIONS = ('core',
            'sources',
            'altnames', 'triggers', 'identifier',
            'classification',
            'endangerment', 'hh_ethnologue_comment', 'iso_retirement')

(CORE,
 SOURCES,
 ALTNAMES, TRIGGERS, IDENTIFIER,
 CLASSIFICATION,
 ENGANGERMENT, HH_ETHNOLOGUE_COMMENT, ISO_RETIREMENT) = SECTIONS

ALL_OPTIONS = object()

FIELDS = {(CORE, 'name'): False,
          (CORE, 'hid'): False,
          (CORE, 'level'): False,
          (CORE, 'iso639-3'): False,
          (CORE, 'latitude'): False,
          (CORE, 'longitude'): False,

          (CORE, 'macroareas'): True,
          (CORE, 'countries'): True,

          (CORE, 'name_comment'): False,

          # FIXME: core hapaxes
          (CORE, 'comment'): False,
          (CORE, 'location'): False,
          (CORE, 'name_pronunciation'): False,
          (CORE, 'speakers'): False,

          (CORE, 'links'): True,
          (CORE, 'timespan'): False,

          (SOURCES, ALL_OPTIONS): True,

          (ALTNAMES, ALL_OPTIONS): True,

          (TRIGGERS, ALL_OPTIONS): True,

          (IDENTIFIER, ALL_OPTIONS): False,

          (CLASSIFICATION, 'sub'): False,
          (CLASSIFICATION, 'subrefs'): True,
          (CLASSIFICATION, 'family'): False,
          (CLASSIFICATION, 'familyrefs'): True,

          (ENGANGERMENT, 'status'): False,
          (ENGANGERMENT, 'source'): False,
          (ENGANGERMENT, 'date'): False,
          (ENGANGERMENT, 'comment'): False,

          (HH_ETHNOLOGUE_COMMENT, 'isohid'): False,
          (HH_ETHNOLOGUE_COMMENT, 'comment_type'): False,
          (HH_ETHNOLOGUE_COMMENT, 'ethnologue_versions'): False,
          (HH_ETHNOLOGUE_COMMENT, 'comment'): False,

          (ISO_RETIREMENT, 'code'): False,
          (ISO_RETIREMENT, 'name'): False,
          (ISO_RETIREMENT, 'change_request'): False,
          (ISO_RETIREMENT, 'effective'): False,
          (ISO_RETIREMENT, 'reason'): False,
          (ISO_RETIREMENT, 'change_to'): True,
          (ISO_RETIREMENT, 'remedy'): False,
          (ISO_RETIREMENT, 'comment'): False}

SECTION_ORDER = [s for s, _ in _tools.groupby_itemgetter(0)(FIELDS)]
SECTION_ORDER = _tools.Ordering.fromlist(SECTION_ORDER, start_index=1)

FIELD_ORDER = _tools.Ordering.fromlist(FIELDS, start_index=1)

CORE_SECTIONS = frozenset({CORE})

# avoid writing empty 'core', 'timespan' options to files (first added in Glottolog 4.2, treedb 0.11)
OMIT_EMPTY_CORE_OPTIONS = frozenset({'timespan'})

KEEP_EMPTY_SECTIONS = CORE_SECTIONS | frozenset({SOURCES})

KEEP_EMPTY_OPTIONS = frozenset({(CORE, 'countries'),
                                (SOURCES, 'glottolog'),
                                (ISO_RETIREMENT, 'change_to')})


log = logging.getLogger(__name__)


def is_known(section, option, /):
    """Retun True if the section option is known or in an ALL_OPTIONS section."""
    return (section, ALL_OPTIONS) in FIELDS or (section, option) in FIELDS


def is_all_options(section, option, /):
    """Retun True if the section option is in an ALL_OPTIONS section."""
    return (section, ALL_OPTIONS) in FIELDS


def is_lines(section, option, /, *, unknown_as_scalar: bool = True):
    """Return True if the section option is treated as list of lines."""
    result = FIELDS.get((section, ALL_OPTIONS))

    if result is None:
        try:
            return FIELDS[section, option]
        except KeyError:
            warnings.warn(f'section {section!r} unknown option {option!r}')

            if unknown_as_scalar:
                log.warning('treating %r as scalar', (section, option))
                return None

            log.exception('unknown option')
            raise

    return result


sorted_sections = SECTION_ORDER.sorted


def sorted_options(section, options, /):
    """Return the given section options as sorted list in canonical order."""
    fields = FIELD_ORDER.sorted((section, o) for o in options)
    return [o for _, o in fields]


def parse_lines(value, /):
    r"""

    >>> parse_lines(None)
    []

    >>> parse_lines(' spam\neggs\n  ')
    ['spam', 'eggs']
    """
    if value is None:
        return []
    return value.strip().splitlines()


def format_lines(value, /):
    r"""

    >>> format_lines(['spam', 'eggs'])
    '\nspam\neggs'

    >>> format_lines([])
    ''
    """
    lines = [''] + value
    return '\n'.join(lines)


RawRecordType: TypeAlias = Mapping[str, Mapping[str, str]]

RawRecordItem: TypeAlias = tuple[_globals.PathType | None, RawRecordType]


def join_lines_inplace(record_item: _globals.RecordItem, /) -> RawRecordItem:
    path, record = record_item
    for name, section in record.items():
        for option in section:
            if is_lines(name, option):
                lines = format_lines(section[option])
                section[option] = lines
    return path, record



import io
import json

import pytest

from helpers import (pairwise,
                     get_assert_head,
                     assert_nonempty_string,
                     assert_nonempty_dict,
                     assert_file_size_between,
                     assert_valid_languoids)

PREFIX = 'path_languoid:path:sha256:'

PREFIX_ID = 'path_languoid:id:sha256:'

CHECKSUM = {'master': None,
            'v5.3': '82d08f00d3596d437a66084f1d97fb68608df4a72d93278a8a8442e053e795ba',
            'v5.2.1': '0f0b5cb36fa2b9c1c8ed087f8bff0e46b7fc37f0c6ad14fc65139486db88c3dd',
            'v5.2': 'c80bc501cf1bef9f3f9449ffa8a6a60d0a4082cfa74f82768490e698dc3e3c70',
            'v5.1': 'e37d43e99deb5fb5055916b29c009c43fafffc2c60c4e9fc027b34d16336d856',
            'v5.0': '41ad0e4b1ef0569c9699f1591f1a3fa9de64bee17cfc8b968330d96eff4d3d84',
            'v4.8': '2a5e60d586fb4c9d6e88efbcc31f5db72af699c584f7053dda9bf28dd3599cab',
            'v4.7': '69f3058e79d443336ad0f33b71ff184100283502a9b52805a1815852df704315',
            'v4.6': '9f63edacc4af792106bcc5a49b6339f0c90e096829c8409b361a6ca2e99107c1',
            'v4.5': '24c79cb12d016ebde9896f3c946b42bebd500318581a41ac17f222a0b292dc9d',
            'v4.4': '4ad1d84119582f24ca9403e4f9a501f1aeb947e1b8d7ea57fad3ef5e4bccdb4c',
            'v4.3-treedb-fixes': '54b468c7310fdd958b2b17fe439ee47c00d211498b405a5bd74b2920f68e3969',
            'v4.2.1': '9e19d66c95a43f595a8ea0b72ba6e7e293e02faf66978b63dda8ddba7d37e3f6',
            'v4.1': '9b795566bd7f5ccb10e0cb4f5e5be10b5ccce496d9816728b904f41b75cdd55a'}

STATS = {'master': None,
         'v5.3': '''\
27,177 languoids
   246 families
   183 isolates
   429 roots
 8,618 languages
 4,607 subfamilies
13,706 dialects
 7,674 Spoken L1 Languages
   227 Sign Language
   128 Unclassifiable
    87 Pidgin
    68 Unattested
    31 Artificial Language
     4 Mixed Language
    19 Speech Register
 8,238 All
   380 Bookkeeping
''',
         'v5.2.1': '''\
27,037 languoids
   247 families
   182 isolates
   429 roots
 8,612 languages
 4,585 subfamilies
13,593 dialects
 7,675 Spoken L1 Languages
   227 Sign Language
   120 Unclassifiable
    87 Pidgin
    68 Unattested
    31 Artificial Language
     8 Mixed Language
    15 Speech Register
 8,231 All
   381 Bookkeeping
''',
         'v5.2': '''\
27,034 languoids
   246 families
   184 isolates
   430 roots
 8,612 languages
 4,583 subfamilies
13,593 dialects
 7,675 Spoken L1 Languages
   227 Sign Language
   120 Unclassifiable
    87 Pidgin
    68 Unattested
    31 Artificial Language
     8 Mixed Language
    15 Speech Register
 8,231 All
   381 Bookkeeping
''',
         'v5.1': '''\
26,953 languoids
   247 families
   183 isolates
   430 roots
 8,605 languages
 4,547 subfamilies
13,554 dialects
 7,667 Spoken L1 Languages
   227 Sign Language
   120 Unclassifiable
    87 Pidgin
    68 Unattested
    31 Artificial Language
     8 Mixed Language
    15 Speech Register
 8,223 All
   382 Bookkeeping
''',
         'v5.0': '''\
26,879 languoids
   246 families
   184 isolates
   430 roots
 8,604 languages
 4,522 subfamilies
13,507 dialects
 7,665 Spoken L1 Languages
   227 Sign Language
   120 Unclassifiable
    84 Pidgin
    68 Unattested
    31 Artificial Language
     9 Mixed Language
    15 Speech Register
 8,219 All
   385 Bookkeeping
''',
         'v4.8': '''\
26,669 languoids
   245 families
   184 isolates
   429 roots
 8,595 languages
 4,467 subfamilies
13,362 dialects
 7,654 Spoken L1 Languages
   223 Sign Language
   121 Unclassifiable
    84 Pidgin
    68 Unattested
    31 Artificial Language
     9 Mixed Language
    15 Speech Register
 8,205 All
   390 Bookkeeping
''',
         'v4.7': '''\
26,416 languoids
   245 families
   184 isolates
   429 roots
 8,572 languages
 4,425 subfamilies
13,174 dialects
 7,636 Spoken L1 Languages
   215 Sign Language
   121 Unclassifiable
    84 Pidgin
    68 Unattested
    32 Artificial Language
     9 Mixed Language
    15 Speech Register
 8,180 All
   392 Bookkeeping
''',
         'v4.6': '''\
26,285 languoids
   245 families
   182 isolates
   427 roots
 8,565 languages
 4,388 subfamilies
13,087 dialects
 7,628 Spoken L1 Languages
   215 Sign Language
   121 Unclassifiable
    84 Pidgin
    68 Unattested
    32 Artificial Language
    10 Mixed Language
    15 Speech Register
 8,173 All
   392 Bookkeeping
''',
         'v4.5': '''\
26,101 languoids
   245 families
   182 isolates
   427 roots
 8,549 languages
 4,348 subfamilies
12,959 dialects
 7,616 Spoken L1 Languages
   210 Sign Language
   121 Unclassifiable
    83 Pidgin
    68 Unattested
    31 Artificial Language
    11 Mixed Language
    15 Speech Register
 8,155 All
   394 Bookkeeping
''',
         'v4.4': '''\
25,900 languoids
   245 families
   182 isolates
   427 roots
 8,533 languages
 4,326 subfamilies
12,796 dialects
 7,613 Spoken L1 Languages
   202 Sign Language
   122 Unclassifiable
    83 Pidgin
    68 Unattested
    31 Artificial Language
    11 Mixed Language
    14 Speech Register
 8,144 All
   389 Bookkeeping
''',
         'v4.3-treedb-fixes': '''\
25,439 languoids
   244 families
   181 isolates
   425 roots
 8,516 languages
 4,265 subfamilies
12,414 dialects
 7,606 Spoken L1 Languages
   196 Sign Language
   123 Unclassifiable
    81 Pidgin
    67 Unattested
    29 Artificial Language
    12 Mixed Language
    10 Speech Register
 8,124 All
   392 Bookkeeping
''',
         'v4.2.1': '''\
25,228 languoids
   242 families
   187 isolates
   429 roots
 8,515 languages
 4,234 subfamilies
12,237 dialects
 7,604 Spoken L1 Languages
   196 Sign Language
   123 Unclassifiable
    81 Pidgin
    66 Unattested
    28 Artificial Language
    14 Mixed Language
    10 Speech Register
 8,122 All
   393 Bookkeeping
''',
        'v4.1': '''\
24,701 languoids
   241 families
   188 isolates
   429 roots
 8,506 languages
 4,170 subfamilies
11,784 dialects
 7,596 Spoken L1 Languages
   194 Sign Language
   122 Unclassifiable
    80 Pidgin
    67 Unattested
    28 Artificial Language
    14 Mixed Language
    10 Speech Register
 8,111 All
   395 Bookkeeping
'''}


def test_print_languoid_stats(pytestconfig, capsys, treedb):
    expected = STATS.get(pytestconfig.option.glottolog_tag)

    assert treedb.print_languoid_stats() is None

    out, err = capsys.readouterr()
    assert not err

    if expected is None:
        assert out.strip()
    else:
        assert out == expected


def test_iterlanguoids(bare_treedb, n=100):
    items = bare_treedb.iterlanguoids()
    assert_valid_languoids(items, n=n)


@pytest.mark.parametrize(
    'source',
    ['files',
     pytest.param('raw', marks=pytest.mark.raw),
     'tables'],
    ids=lambda x: f'source={x}')
def test_checksum(pytestconfig, treedb, source):
    expected = CHECKSUM.get(pytestconfig.option.glottolog_tag)

    result = treedb.checksum()

    if expected is None:
        assert result.startswith(PREFIX)
        assert len(result) - len(PREFIX) == 64
    else:
        assert result == PREFIX + expected


def test_write_json_lines_checksum(pytestconfig, treedb):
    expected = CHECKSUM.get(pytestconfig.option.glottolog_tag)

    with io.StringIO() as buf:
        treedb.write_languoids(buf)
        value = buf.getvalue()

    result = treedb.sha256sum(value, hash_file_string=True)

    if expected is None:
        assert len(result) == 64
    else:
        assert result == expected


@pytest.mark.parametrize('suffix',
                         ['.jsonl', '.jsonl.gz'],
                         ids=lambda x: f'suffix={x}')
def test_write_json_lines(pytestconfig, capsys, treedb, suffix, n=100):
    name_suffix = '-memory' if treedb.engine.file is None else ''
    name_suffix += pytestconfig.option.file_engine_tag
    args = ([f'treedb{name_suffix}.languoids{suffix}'] if suffix != 'jsonl.gz'
            else [])

    filepath, _ = treedb.write_languoids(*args)

    assert filepath.name == f'treedb{name_suffix}.languoids{suffix}'
    assert_file_size_between(filepath, 1, 200)

    if filepath.name.endswith('.jsonl'):
        with filepath.open(encoding='utf-8') as f:
            for line in get_assert_head(f, n=n):
                item = json.loads(line)
                assert_nonempty_dict(item)

                path = item['__path__']
                assert isinstance(path, list)
                assert all(isinstance(p, str) for p in path)
                assert path
                assert all(path)

                languoid = item['languoid']
                assert_nonempty_dict(languoid)

                for key in ('id', 'level', 'name'):
                    assert_nonempty_string(languoid[key])

                assert languoid['parent_id'] or languoid['parent_id'] is None
                assert languoid['level'] in ('family', 'language', 'dialect')

    out, err = capsys.readouterr()
    assert not out
    assert not err

    expected_checksum = CHECKSUM.get(pytestconfig.option.glottolog_tag)
    if expected_checksum is not None:
        assert treedb.sha256sum(filepath) == expected_checksum


def test_write_json_lines_indent(tmp_path, treedb, limit=2):
    target = tmp_path / 'languoids.json.txt'
    with pytest.warns(UserWarning,
                      match=r'non-canonical JSON Lines format from indent 2'):
        path, n_written = treedb.write_languoids(target, pretty=True,
                                                 limit=limit)

    assert path == target
    assert n_written

    text = path.read_text(encoding='utf-8')
    assert text.startswith('{\n  "__path__": [')
    assert text.endswith('\n    }\n  }\n}\n')


@pytest.mark.pandas
@pytest.mark.parametrize(
    'source',
    ['files',
     pytest.param('raw', marks=pytest.mark.raw),
     'tables'],
    ids=lambda x: f'source={x}')
def test_pd_read_languoids(treedb, source, limit=1_000):
    df = treedb.pd_read_languoids(source=source)

    if treedb.backend.pandas.PANDAS is None:
        assert df is None
    else:
        assert not df.empty
        assert df.index.name == 'id'
        assert list(df.columns) == ['path', 'languoid']
        assert df.index.is_unique
        df.info(memory_usage='deep')


xfail_master_unnormalized = pytest.mark.xfail_glottolog_tag(
    'master', reason='possibly unnormalized master',
    raises=AssertionError)


xfail_4_8_empty_altnames_elcat = pytest.mark.xfail_glottolog_tag(
    'v4.8', reason='empty v4.8 altnames',
    # https://github.com/glottolog/glottolog/pull/980
    raises=AssertionError)


xfail_4_5_empty_altnames_elcat = pytest.mark.xfail_glottolog_tag(
    'v4.5', reason='empty v4.4 altnames',
    # https://github.com/glottolog/glottolog/pull/795
    raises=AssertionError)


xfail_4_1_float_normalization = pytest.mark.xfail_glottolog_tag(
    'v4.1', reason='missing v4.1 float normalization',
    # https://github.com/glottolog/glottolog/pull/495
    raises=AssertionError)


@pytest.mark.parametrize(
    'kwargs',
    [pytest.param([{'source': 'files'},
                   {'source': 'raw'},
                   {'source': 'tables'}],
                  id='files, raw, tables',
                  marks=[xfail_master_unnormalized,
                         xfail_4_8_empty_altnames_elcat,
                         xfail_4_5_empty_altnames_elcat,
                         xfail_4_1_float_normalization]),
     pytest.param([{'source': 'raw', 'order_by': 'id'},
                   {'source': 'tables', 'order_by': 'id'}],
                  id='raw(order_by=id), tables(order_by=id)',
                  marks=[xfail_master_unnormalized,
                         xfail_4_8_empty_altnames_elcat,
                         xfail_4_1_float_normalization]),
     pytest.param([{'source': 'files', 'limit': 100,
                    'expected_prefix': 'path_languoid:path[limit=100]:sha256:'},
                   {'source': 'raw', 'order_by': 'file', 'limit': 100,
                    'expected_prefix': 'path_languoid:file[limit=100]:sha256:'}],
                  id='files(limit=100), raw(order_by=file, limit=100)',
                  marks=[xfail_4_5_empty_altnames_elcat])])
def test_checksum_equivalence(pytestconfig, treedb, kwargs):
    """Test for equivalence of the serialization from different sources.

    - from the md.ini files ('files')
    - from the file lines loaded into raw tables  ('raw')
    - from the parsed table contents ('tables')

    This should pass for glottolog release candidates.

    This is intended to pass at glottolog HEAD with the latest treedb.

    This can fail from manual editing inconsistencies.
    Running treedb.write_files() can be used to write back the table contents
    to the glottolog md.ini files in a soft manner (e.g. not normalizing order
    for cases where it is not significant). If the diff looks correct, this
    can be used to bring the repository files back into a canonicalized
    (i.e. round-tripable) format. A common instance are floats being normalized
    via round-tripping (e.g. dropping trailing zeros).

    Changing the export formatting breaks this intentionally for older versions,
    i.e. for glottolog commits before the change was applied to the
    glottolog repository md.ini files (see xfails above).

    Intended changes usually pertain to normalization/canonicalization,
    ordering, formatting.

    Less frequent changes pertain to changing the data format such as
    removing redundancy.

    --
    First shalt thou take out the Holy Pin.

    Then shalt thou count to three, no more, no less.

    Three shall be the number thou shalt count,
    and the number of the counting shall be three.

    Four shalt thou not count, neither count thou two,
    excepting that thou then proceed to three.

    Five is right out.

    Once the number three, being the third number, be reached,
    then lobbest thou thy Holy Hand Grenade of Antioch towards thy foe,
    who, being naughty in My sight, shall snuff it.
    """
    def iterchecksums(kwargs):
        for kw in kwargs:
            expected_prefix = kw.pop('expected_prefix', None)
            # --exclude-raw: filter out raw sources instead of skipping
            if pytestconfig.option.exclude_raw and kw.get('source') == 'raw':
                continue
            checksum = treedb.checksum(**kw)
            prefix, colon, hexdigest = checksum.rpartition(':')
            prefix += colon
            yield kw, expected_prefix, prefix, hexdigest

    results = list(iterchecksums(kwargs))

    for kw, expected_prefix, prefix, hexdigest in results:
        assert len(hexdigest) == 64
        if expected_prefix is None:
            expected_prefix = PREFIX_ID if kw.get('order_by') == 'id' else PREFIX
        assert prefix == expected_prefix

    for (c, _, _, cur), (n, _, _, nxt) in pairwise(results):
        cur, nxt = (s.rpartition(':')[2] for s in (cur, nxt))
        assert cur == nxt, f'checksum(**{c!r}) == checksum(**{n!r})'



"""Yield languoids, write information to stdout, .csv, .jsonl, etc."""

from collections.abc import Iterable
import datetime
import itertools
import hashlib
import logging
import operator
import warnings

from . import _globals
from . import _tools
from . import backend as _backend
from .backend import export as _backend_export
from .backend import pandas as _backend_pandas
from .languoids import records as _records
from .models import SPECIAL_FAMILIES, BOOKKEEPING
from . import queries as _queries

__all__ = ['print_languoid_stats',
           'iterlanguoids',
           'checksum',
           'write_json_lines',
           'pd_read_languoids',
           'fetch_languoids',
           'write_files']

CHECKSUM_NAME = 'path_languoid'

FALLBACK_ENGINE_PATH = _tools.path_from_filename('treedb.sqlite3')


log = logging.getLogger(__name__)


def print_languoid_stats(*, file=None,
                         bind=_globals.ENGINE):
    select_stats = _queries.get_stats_query()
    rows = _backend.iterrows(select_stats, mappings=True, bind=bind)
    rows, counts = itertools.tee(rows)

    _backend_export.print_rows(rows, format_='{n:6,d} {kind}',
                               file=file, bind=None)

    sums = [('languoids', ('families', 'languages', 'subfamilies', 'dialects')),
            ('roots', ('families', 'isolates')),
            ('All', ('Spoken L1 Languages',) + SPECIAL_FAMILIES),
            ('languages', ('All', BOOKKEEPING))]

    counts = {c['kind']: c['n'] for c in counts}
    for total, parts in sums:
        values = [counts[p] for p in parts]
        parts_sum = sum(values)
        term = ' + '.join(f'{v:,d} {p}' for p, v in zip(parts, values))
        log.debug('verify %s == %d %s', term, counts[total], total)
        if counts[total] != parts_sum:  # pragma: no cover
            warnings.warn(f'{term} = {parts_sum:,d}')


def iterlanguoids(source: str = 'files', /, *,
                  limit: int | None = None,
                  offset: int | None = 0,
                  order_by: str = _globals.LANGUOID_ORDER,
                  progress_after: int = _tools.PROGRESS_AFTER,
                  root=_globals.ROOT, bind=_globals.ENGINE,
                  ) -> Iterable[_globals.LanguoidItem]:
    """Yield (path, languoid) pairs from diffferent sources."""
    log.info('generate languoids from %r', source)
    if source in ('files', 'raw'):
        log.info('extract languoids from %r', source)
        if source == 'files':
            if order_by not in ('path', 'file', True, None, False):  # pragma: no cover
                raise ValueError(f'order_by={order_by!r} not implemented')
            else:
                log.debug('rely on %r as %s order', order_by, source)
                del order_by

            from . import languoids

            records = languoids.iterrecords(root=root, progress_after=progress_after)
        elif source == 'raw':
            from . import raw

            records = raw.fetch_records(order_by=order_by,
                                        progress_after=progress_after,
                                        bind=bind)

        items = _records.pipe(records, dump=False,
                              convert_lines=(source == 'files'))
        return _tools.islice_limit(items,
                                   limit=limit,
                                   offset=offset)
    elif source == 'tables':
        return fetch_languoids(limit=limit,
                               offset=offset,
                               order_by=order_by,
                               progress_after=progress_after,
                               bind=bind)
    else:  # pragma: no cover
        raise ValueError(f'unknown source: {source!r}')


def checksum(source: str = 'tables', *,
             limit: int | None = None,
             offset: int | None = 0,
             order_by: str = _globals.LANGUOID_ORDER,
             hash_name: str = _globals.DEFAULT_HASH,
             bind=_globals.ENGINE):
    """Return checksum over source."""
    log.info('hash languoids json lines from %r ordered by %r with %r',
             source, order_by, hash_name)
    hashobj = hashlib.new(hash_name)
    assert hasattr(hashobj, 'hexdigest')
    hashobj, total_lines = write_json_lines(hashobj,
                                            source=source,
                                            limit=limit,
                                            offset=offset,
                                            order_by=order_by,
                                            sort_keys=True, bind=bind)
    log.info('%s json lines written', f'{total_lines:_d}')

    offset = f'{offset=!r}' if offset else ''
    limit = f'{limit=!r}' if limit is not None else ''
    sliced = ','.join(s for s in (offset, limit) if s)
    sliced = f'[{sliced}]' if sliced else ''
    result = (f'{CHECKSUM_NAME}:{order_by}{sliced}'
              f':{hashobj.name}:{hashobj.hexdigest()}')
    log.info('%s: %r', hashobj.name, result)
    return result


def write_json_lines(file=None, /, *, suffix: str = '.jsonl',
                     delete_present: bool = True,
                     autocompress: bool = True,
                     source: str = 'tables',
                     limit: int | None = None,
                     offset: int | None = 0,
                     order_by: str = _globals.LANGUOID_ORDER,
                     sort_keys: bool = True,
                     pretty: bool = False,
                     ensure_ascii: bool = False,
                     path_label: str = _globals.PATH_LABEL,
                     languoid_label: str = _globals.LANGUOID_LABEL,
                     bind=_globals.ENGINE):
    r"""Write languoids as newline delimited JSON.

    $ python -c "import sys, treedb; treedb.load('treedb.sqlite3'); treedb.write_json_lines(sys.stdout)" \
    | jq -s "group_by(.languoid.level)[]| {level: .[0].languoid.level, n: length}"

    $ jq "del(recurse | select(. == null or arrays and empty))" treedb.languoids.jsonl > treedb.languoids-jq.jsonl
    """  # noqa: E501
    if file is None:
        file = FALLBACK_ENGINE_PATH if source == 'files' else bind.file
        file = file.with_name(f'{file.stem}-{source}.languoids{suffix}')

    log.info('write json lines from %s to %r', source, file)
    pipe_kwargs = {'sort_keys': sort_keys,
                   'compact': True,
                   'indent': None,
                   'ensure_ascii': ensure_ascii,
                   'delete_present': delete_present,
                   'autocompress': autocompress,
                   'newline': '\n'}
    if pretty:
        pipe_kwargs.update(compact=False, indent=2)
    if source in ('files', 'raw'):
        items = iterlanguoids(source,
                              limit=limit, offset=offset,
                              order_by=order_by, bind=bind)
        items = ({path_label: path, languoid_label: languoid}
                 for path, languoid in items)
        return _tools.pipe_json_lines(file, items, **pipe_kwargs)
    elif source == 'tables':
        sqlite_format = not pretty and not ensure_ascii
        query = _queries.get_json_query(limit=limit,
                                        offset=offset,
                                        as_rows=False,
                                        load_json=not sqlite_format,
                                        order_by=order_by,
                                        sort_keys=sort_keys,
                                        path_label=path_label,
                                        languoid_label=languoid_label)
        if sqlite_format:  # fast path
            log.debug('use raw SQLite JSON format string')
            pipe_func = _tools.pipe_lines
            for kw in ('sort_keys', 'compact', 'indent', 'ensure_ascii'):
                pipe_kwargs.pop(kw)
        else:
            log.info('roundtrip SQLite JSON for reformatting')
            pipe_func = _tools.pipe_json_lines

        with _backend.connect(bind=bind) as conn:
            lines = conn.execute(query).scalars()
            result = pipe_func(file, lines, **pipe_kwargs)
        return result
    else:  # pragma: no cover
        raise ValueError(f'unknown source: {source!r}')


def pd_read_languoids(*, source: str = 'tables',
                      limit: int | None = None,
                      offset: int | None = 0,
                      order_by: str = _globals.LANGUOID_ORDER,
                      sort_keys: bool = True,
                      path_label: str = _globals.PATH_LABEL,
                      languoid_label: str = _globals.LANGUOID_LABEL,
                      bind=_globals.ENGINE,
                      **kwargs):
    log.info('read json lines with pd.read_json(lines=True)')
    if source in ('files', 'raw'):
        items = iterlanguoids(source,
                              limit=limit,
                              offset=offset,
                              order_by=order_by,
                              bind=bind)
        items = ({path_label: path, languoid_label: languoid}
                 for path, languoid in items)
        json_lines = _tools.pipe_json(items, dump=True)
        df = _backend_pandas._pd_read_json_lines(json_lines,
                                                 orient='record',
                                                 **kwargs)
    elif source == 'tables':
        query = _queries.get_json_query(limit=limit,
                                        offset=offset,
                                        as_rows=False,
                                        load_json=False,
                                        order_by=order_by,
                                        sort_keys=sort_keys,
                                        path_label=path_label,
                                        languoid_label=languoid_label)

        df = _backend_pandas.pd_read_json_lines(query,
                                                orient='record',
                                                bind=bind,
                                                **kwargs)
    if df is not None:
        df.rename(columns={_globals.PATH_LABEL: 'path'}, inplace=True)
        index = df['languoid'].map(operator.itemgetter('id')).rename('id')
        df.set_index(index, inplace=True, verify_integrity=True)
    return df


def fetch_languoids(*, limit: int | None = None,
                    offset: int | None = 0,
                    order_by: str = _globals.LANGUOID_ORDER,
                    progress_after: int = _tools.PROGRESS_AFTER,
                    bind=_globals.ENGINE):
    log.info('fetch languoids from json query, order_by: %r', order_by)
    query = _queries.get_json_query(limit=limit,
                                    offset=offset,
                                    order_by=order_by,
                                    as_rows=True,
                                    load_json=True)
    del limit, offset

    json_datetime = datetime.datetime.fromisoformat

    rows = _backend.iterrows(query, bind=bind)

    n = 0
    make_item = _globals.LanguoidItem.from_filepath_languoid
    for n, (path, item) in enumerate(rows, start=1):
        endangerment = item['endangerment']
        if endangerment is not None:
            endangerment['date'] = json_datetime(endangerment['date'])
        if not item.get('timespan'):
            item.pop('timespan', None)
        yield make_item(path, item)

        if not (n % progress_after):
            log.info('%s languoids fetched', f'{n:_d}')

    log.info('%s languoids total', f'{n:_d}')


def write_files(root=_globals.ROOT, /, *, replace: bool = False,
                dry_run: bool = False,
                require_nwritten: int | None = None,
                source: str = 'tables',
                limit: int | None = None,
                offset: int | None = 0,
                progress_after: int = _tools.PROGRESS_AFTER,
                bind=_globals.ENGINE) -> int:
    log.info('write from %r to tree %r', source, root)
    if source == 'raw_lines':
        from . import raw

        return raw.write_files(root, replace=replace,
                               dry_run=dry_run,
                               require_nwritten=require_nwritten,
                               limit=limit,
                               offset=offset,
                               progress_after=progress_after,
                               bind=bind)

    from .languoids import files

    languoids = iterlanguoids(source,
                              limit=limit,
                              offset=offset,
                              order_by='path',
                              progress_after=progress_after,
                              bind=bind)

    records = _records.pipe(languoids, dump=True, convert_lines=False)

    return files.write_files(records, root=root, replace=replace,
                             dry_run=dry_run,
                             require_nwritten=require_nwritten,
                             progress_after=progress_after)
